
import re

from django.views.generic import View as DjangoGenericView
from django.db import transaction
from django.db.utils import DatabaseError
from django.core.exceptions import (
    ObjectDoesNotExist,
    MultipleObjectsReturned,
    ValidationError,
)

from lily.conf import settings
from .events import EventFactory
from . import serializers
from .utils import import_from_string
from .access import Access
from .context import Context
from .source import Source
from .input import Input
from .output import Output
from .name import ConstantName


event = EventFactory()


class command_override:  # noqa

    def __init__(
            self,
            name,
            meta=None,
            access=None,
            input=None,
            output=None,
            is_atomic=None):

        self.name = name
        self.meta = meta
        self.access = access
        self.input = input
        self.output = output
        self.is_atomic = is_atomic


class HTTPCommands(DjangoGenericView):

    @classmethod
    def overwrite(cls, get=None, post=None, put=None, delete=None):

        class cls_copy(cls):  # noqa
            pass

        for method_name, verb in [
                ('get', get),
                ('post', post),
                ('put', put),
                ('delete', delete)]:

            method = getattr(cls_copy, method_name, None)
            if method and verb:
                conf = method.command_conf

                setattr(
                    cls_copy,
                    method_name,
                    command(
                        name=verb.name,
                        meta=verb.meta or conf['meta'],
                        access=verb.access or conf['access'],
                        input=verb.input or conf['input'],
                        output=verb.output or conf['output'],
                        is_atomic=(
                            verb.is_atomic is not None and verb.is_atomic or
                            conf['is_atomic']),
                    )(conf['fn']))

        return cls_copy


def command(
        name,
        meta,
        access=None,
        input=None,
        output=None,
        is_atomic=False):

    # -- defaults
    name = (isinstance(name, str) and ConstantName(name)) or name
    access = access or Access(access_list=None)
    input = input or Input()
    output = output or Output(serializer=serializers.EmptySerializer)

    def command_inner(fn):

        def inner(self, request, *args, **kwargs):

            self.event = event

            request._lily_context = Context(
                command_name=name.render_command_name(),
                request=request)

            try:
                #
                # AUTHORIZATION
                #
                if access.access_list:
                    authorizer = import_from_string(
                        settings.LILY_AUTHORIZER_CLASS
                    )(access.access_list)
                    request.access = authorizer.authorize(request)
                    request.log_authorizer = authorizer.log(request.access)

                #
                # INPUT
                #
                if input:
                    input.parse(
                        request,
                        command_name=request._lily_context.command_name)

                #
                # IS_ATOMIC
                #
                if not is_atomic:
                    # -- trigger the handler which should raise the success or
                    # -- error response
                    # -- if no exception will be raised we return whatever it
                    # -- was returned since most likely some default Django's
                    # -- mechanism took place
                    return fn(self, request, *args, **kwargs)

                else:
                    success_exception = None
                    with transaction.atomic(using=is_atomic):
                        try:
                            return fn(self, request, *args, **kwargs)

                        except EventFactory.BaseSuccessException as e:
                            e.extend(
                                context=request,
                                event=name.render_event_name(request, e)).log()

                            success_exception = e

                    # -- re-raise success exceptions outside of the
                    # -- transaction block in order to make sure that it
                    # -- would not be interpreted by `atomic` block as an
                    # -- error
                    if success_exception:
                        raise success_exception

            except EventFactory.Generic as e:
                return e.extend(
                    method=request.method, path=request.path
                ).log().response()

            except EventFactory.BaseSuccessException as e:
                e.extend(
                    context=request,
                    event=name.render_event_name(request, e)).log()

                #
                # RESPONSE VALIDATION (Test Server Only)
                #
                if e.data and request.META.get('SERVER_NAME') == 'testserver':
                    serializer = output.serializer(
                        data=e.data,
                        context={
                            **e.output_context,
                            'request': request,
                            'command_name': request._lily_context.command_name,
                        })

                    if not serializer.is_valid():
                        e = event.BrokenRequest(
                            'RESPONSE_DID_NOT_VALIDATE',
                            context=request,
                            data={'errors': serializer.errors},
                            is_critical=True)

                        return e.response_class(e.data)

                #
                # OUTPUT
                #
                try:
                    body = output.serializer(
                        (
                            (e.data is not None and e.data) or
                            (e.instance is not None and e.instance)
                        ),
                        context={
                            **e.output_context,
                            'request': request,
                            'command_name': request._lily_context.command_name,
                        }).data

                    body['@event'] = e.event

                    return e.response_class(body)

                # -- case of serializer returning error as well
                except EventFactory.BaseErrorException as e:
                    e.update_with_context(context=request)

                    response = e.response_class(e.data)
                    if e.extra_headers:
                        for k, v in e.extra_headers.items():
                            response[k] = v

                    return response

            except EventFactory.BaseErrorException as e:
                e.update_with_context(context=request)

                response = e.response_class(e.data)
                if e.extra_headers:
                    for k, v in e.extra_headers.items():
                        response[k] = v

                return response

            #
            # GENERIC ERRORS HANDLING
            #
            except ValidationError as e:
                e = event.BrokenRequest(
                    'BODY_JSON_DID_NOT_PARSE',
                    context=request,
                    data={'errors': e.message_dict})
                return e.response_class(e.data)

            except ObjectDoesNotExist as e:
                # -- Rather Hacky way of fetching the name of model
                # -- which raised the DoesNotExist error
                model_name = str(e).split()[0].upper()
                e = event.DoesNotExist(
                    'COULD_NOT_FIND_{}'.format(model_name),
                    context=request)
                return e.response_class(e.data)

            except MultipleObjectsReturned as e:
                # -- Rather Hacky way of fetching the name of model
                # -- which raised the MultipleObjectsReturned error
                m = re.search(r'than one\s+(?P<model_name>\w+)', str(e))
                model_name = m.group('model_name').upper()

                e = event.ServerError(
                    'FOUND_MULTIPLE_INSTANCES_OF_{}'.format(model_name),
                    context=request,
                    is_critical=True)
                return e.response_class(e.data)

            # -- handle gracefully database and generic exceptions
            # -- to make sure that lily valid response will be generated
            except DatabaseError:
                e = event.ServerError(
                    'DATABASE_ERROR_OCCURRED',
                    context=request,
                    is_critical=True)

                return e.response_class(e.data)

            except Exception as err:
                e = event.ServerError(
                    'GENERIC_ERROR_OCCURRED',
                    context=request,
                    data={'errors': [str(err)]},
                    is_critical=True)

                return e.response_class(e.data)

        # -- the below specs are available shortly after the code compilation
        # -- and therefore can be made available on runtime
        inner.command_conf = {
            # -- command input
            'meta': meta,
            'access': access,
            'input': input,
            'output': output,
            'is_atomic': is_atomic,
            'fn': fn,
            # -- derived values
            'name': name.render_command_name(),
            'method': fn.__name__,
            'source': Source(fn),
        }

        return inner

    return command_inner




from copy import deepcopy
from unittest.mock import Mock, call

from django.test import TestCase
import pytest

from lily.docs.renderers.angular.command import Command
from lily.base.utils import normalize_indentation
from tests import remove_white_chars


CONF = {
    'method': 'POST',
    'meta': {
        'title': 'Create Path',
        'domain': {
            'id': 'paths',
            'name': 'Paths Management',
        }
    },
    'access': {
        'is_private': True,
        'access_list': ['LEARNER'],
    },
    'path_conf': {
        'path': '/paths/',
        'parameters': []
    },
    'schemas': {
        'input_query': None,
        'input_body': None,
        'output': {'schema': {'hi': 'there'}, 'uri': ''},
    },
    'examples': {},
}


class CommandTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker):
        self.mocker = mocker

    def test_properties(self):

        command_signature = Mock()
        signature = self.mocker.patch(
            'lily.docs.renderers.angular.command.Signature')
        signature.return_value = command_signature
        (
            request_query,
            request_body,
            response,
            request_query_interface,
            request_body_interface,
            response_interface,
        ) = Mock(), Mock(), Mock(), Mock(), Mock(), Mock()
        interface = self.mocker.patch(
            'lily.docs.renderers.angular.command.Interface')
        interface.TYPES.REQUEST_QUERY = 'REQUEST_QUERY'
        interface.TYPES.REQUEST_BODY = 'REQUEST_BODY'
        interface.TYPES.RESPONSE = 'RESPONSE'
        interface.side_effect = [
            request_query_interface,
            request_body_interface,
            response_interface,
        ]
        self.mocker.patch.object(
            Command, 'get_bulk_read_field').return_value = None
        command = Command(
            'READ_TASK',
            {
                'method': 'POST',
                'meta': {
                    'title': 'Read Task',
                    'description': 'Read Task Now',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks/{task_id}',
                    'parameters': [
                        {
                            'name': 'task_id',
                            'type': 'integer',
                        }
                    ]
                },
                'schemas': {
                    'input_query': {'schema': request_query, 'uri': 'Q'},
                    'input_body': {'schema': request_body, 'uri': 'B'},
                    'output': {'schema': response, 'uri': 'R'},
                },
                'examples': {},
            })

        assert command.method == 'post'
        assert command.camel_name == 'readTask'
        assert command.domain_id == 'tasks'
        assert command.domain_name == 'Tasks Management'
        assert command.title == 'Read Task'
        assert command.is_private is True
        assert command.authorization_required is True
        assert command.path == '/tasks/{task_id}'
        assert command.path_parameters == [
            {'name': 'task_id', 'type': 'integer'}]
        assert command.request_query == request_query_interface
        assert command.request_body == request_body_interface
        assert command.response == response_interface
        assert command.signature == command_signature
        assert signature.call_args_list == [call(
            'post',
            '/tasks/{task_id}',
            [{'name': 'task_id', 'type': 'integer'}],
            True,
            request_query_interface,
            request_body_interface,
            bulk_read_field=None,
        )]

        assert interface.call_args_list == [
            call('READ_TASK', 'REQUEST_QUERY', request_query, 'Q'),
            call('READ_TASK', 'REQUEST_BODY', request_body, 'B'),
            call('READ_TASK', 'RESPONSE', response, 'R', bulk_read_field=None),
        ]

    def test_header(self):
        # -- with title and description
        command = Command(
            'READ_TASK',
            {
                'method': 'GET',
                'meta': {
                    'title': 'Read Task',
                    'description': 'Read Task Now',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks',
                    'parameters': []
                },
                'schemas': {
                    'input_query': None,
                    'input_body': None,
                    'output': {'schema': {}, 'uri': ''},
                },
                'examples': {},
            })

        assert command.header == normalize_indentation('''
            /**
             * Read Task
             */
        ''', 0)

        # -- without description
        command = Command(
            'READ_TASK',
            {
                'method': 'GET',
                'meta': {
                    'title': 'Bulk Read Paths',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks',
                    'parameters': []
                },
                'schemas': {
                    'input_query': None,
                    'input_body': None,
                    'output': {'schema': {}, 'uri': ''},
                },
                'examples': {},
            })

        assert command.header == normalize_indentation('''
            /**
             * Bulk Read Paths
             */
        ''', 0)

    def test_render__get_bulk_read_field__not_get(self):

        self.mocker.patch.object(
            Command, 'get_bulk_read_field').return_value = 'people'
        command = Command(
            'UPDATE_TASK',
            {
                'method': 'PUT',
                'meta': {
                    'title': 'Update Task',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks/{task_id}',
                    'parameters': [
                        {
                            'name': 'task_id',
                            'type': 'integer',
                        },
                    ]
                },
                'schemas': {
                    'input_query': None,
                    'input_body': {'schema': {'hi': 'there'}, 'uri': ''},
                    'output': {
                        'schema': {
                            'type': 'object',
                            'required': ['people'],
                            'properties': {
                                'people': {
                                    'type': 'array',
                                },
                            },
                        },
                        'uri': '',
                    },
                },
                'examples': {},
            })

        assert command.render() == normalize_indentation('''
            /**
             * Update Task
             */
            public updateTask(taskId: any, body: X.UpdateTaskBody): Observable<X.UpdateTaskResponse> {
                return this.client
                    .put<X.UpdateTaskResponse>(`/tasks/${taskId}`, body, { authorizationRequired: true })
                    .pipe(map(x => x.people));
            }
        ''', 0)  # noqa

    def test_render__get_bulk_read_field__get(self):

        self.mocker.patch.object(
            Command, 'get_bulk_read_field').return_value = 'people'
        command = Command(
            'READ_TASK',
            {
                'method': 'GET',
                'meta': {
                    'title': 'Read Task',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks/{task_id}',
                    'parameters': [
                        {
                            'name': 'task_id',
                            'type': 'integer',
                        },
                    ]
                },
                'schemas': {
                    'input_query': {'schema': {'hi': 'there'}, 'uri': ''},
                    'input_body': None,
                    'output': {
                        'schema': {
                            'type': 'object',
                            'required': ['people'],
                            'properties': {
                                'people': {
                                    'type': 'array',
                                },
                            },
                        },
                        'uri': '',
                    },
                },
                'examples': {},
            })

        assert command.render() == normalize_indentation('''
            /**
             * Read Task
             */
            public readTask(taskId: any, params: X.ReadTaskQuery): Observable<X.ReadTaskResponse> {
                return this.client.get<X.ReadTaskResponse>(`/tasks/${taskId}`, { params, authorizationRequired: true });
            }
        ''', 0)  # noqa

    def test_render__get(self):

        command = Command(
            'READ_TASK',
            {
                'method': 'GET',
                'meta': {
                    'title': 'Read Task',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks/{task_id}',
                    'parameters': [
                        {
                            'name': 'task_id',
                            'type': 'integer',
                        },
                    ]
                },
                'schemas': {
                    'input_query': {'schema': {'hi': 'there'}, 'uri': ''},
                    'input_body': None,
                    'output': {'schema': {'hi': 'there'}, 'uri': ''},
                },
                'examples': {},
            })

        assert command.render() == normalize_indentation('''
            /**
             * Read Task
             */
            public readTask(taskId: any, params: X.ReadTaskQuery): Observable<X.ReadTaskResponse> {
                return this.client.get<X.ReadTaskResponse>(`/tasks/${taskId}`, { params, authorizationRequired: true });
            }
        ''', 0)  # noqa

    def test_render__post(self):

        command = Command(
            'READ_TASK',
            {
                'method': 'POST',
                'meta': {
                    'title': 'Create Path',
                    'domain': {
                        'id': 'paths',
                        'name': 'Paths Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/paths/',
                    'parameters': []
                },
                'schemas': {
                    'input_query': None,
                    'input_body': {'schema': {'hi': 'there'}, 'uri': ''},
                    'output': {'schema': {'hi': 'there'}, 'uri': ''},
                },
                'examples': {},
            })

        assert command.render() == normalize_indentation('''
            /**
             * Create Path
             */
            public readTask(body: X.ReadTaskBody): Observable<X.ReadTaskResponse> {
                return this.client
                    .post<X.ReadTaskResponse>('/paths/', body, { authorizationRequired: true })
                    .pipe(filter(x => !_.isEmpty(x)));
            }
        ''', 0)  # noqa

    def test_render_facade__get_bulk_read_field(self):

        self.mocker.patch.object(
            Command, 'get_bulk_read_field').return_value = 'people'
        command = Command(
            'READ_TASK',
            {
                'method': 'GET',
                'meta': {
                    'title': 'Read Task',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks/{task_id}',
                    'parameters': [
                        {
                            'name': 'task_id',
                            'type': 'integer',
                        },
                    ]
                },
                'schemas': {
                    'input_query': {'schema': {'hi': 'there'}, 'uri': ''},
                    'input_body': None,
                    'output': {
                        'schema': {
                            'type': 'object',
                            'required': ['people'],
                            'properties': {
                                'people': {
                                    'type': 'array',
                                },
                            },
                        },
                        'uri': '',
                    },
                },
                'examples': {},
            })

        assert command.render_facade() == normalize_indentation('''
            readTask(taskId: any, params: X.ReadTaskQuery): Observable<X.ReadTaskResponse> {
                return this.tasksDomain.readTask(taskId, params);
            }
        ''', 0)  # noqa

    def test_render_facade__get(self):

        command = Command(
            'READ_TASK',
            {
                'method': 'GET',
                'meta': {
                    'title': 'Read Task',
                    'domain': {
                        'id': 'tasks',
                        'name': 'Tasks Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/tasks/{task_id}',
                    'parameters': [
                        {
                            'name': 'task_id',
                            'type': 'integer',
                        },
                    ]
                },
                'schemas': {
                    'input_query': {'schema': {'hi': 'there'}, 'uri': ''},
                    'input_body': None,
                    'output': {'schema': {'hi': 'there'}, 'uri': ''},
                },
                'examples': {},
            })

        assert command.render_facade() == normalize_indentation('''
            readTask(taskId: any, params: X.ReadTaskQuery): Observable<X.ReadTaskResponse> {
                return this.tasksDomain.readTask(taskId, params);
            }
        ''', 0)  # noqa

    def test_render_facade__post(self):

        command = Command(
            'READ_TASK',
            {
                'method': 'POST',
                'meta': {
                    'title': 'Create Path',
                    'domain': {
                        'id': 'paths',
                        'name': 'Paths Management',
                    }
                },
                'access': {
                    'is_private': True,
                    'access_list': ['LEARNER'],
                },
                'path_conf': {
                    'path': '/paths/',
                    'parameters': []
                },
                'schemas': {
                    'input_query': None,
                    'input_body': {'schema': {'hi': 'there'}, 'uri': ''},
                    'output': {'schema': {'hi': 'there'}, 'uri': ''},
                },
                'examples': {},
            })

        assert command.render_facade() == normalize_indentation('''
            readTask(body: X.ReadTaskBody): Observable<X.ReadTaskResponse> {
                return this.pathsDomain.readTask(body);
            }
        ''', 0)  # noqa

    #
    # render_examples
    #
    def test_render_examples(self):

        conf = deepcopy(CONF)
        conf['examples'] = {
            '404 (NOT_FOUND)': {
                'response': {
                    'where': 'here',
                    'error': 'not found',
                    'status': 404,
                },
            },
            '200 (YO)': {
                'response': {
                    'id': 45,
                    'age': 879,
                },
            },

        }

        command = Command('BULK_READ_TASKS', conf)

        assert (
            remove_white_chars(command.render_examples()) ==
            remove_white_chars(normalize_indentation('''
                /**
                 * Examples for BULK_READ_TASKS
                 */
                export const BulkReadTasksExamples = {
                    "200 (YO)": {
                        "age": 879,
                        "id": 45
                    },

                    "404 (NOT_FOUND)": {
                        "error": "not found",
                        "status": 404,
                        "where": "here"
                    }
                }
            ''', 0)))

    #
    # render_access
    #
    def test_render_access(self):

        conf = deepcopy(CONF)
        conf['access'] = {
            'access_list': [
                'LEARNER',
                'MENTOR',
            ],
            'is_private': False,
        }

        command = Command('BULK_READ_TASKS', conf)

        assert command.render_access('AccountType') == (
            'BULK_READ_TASKS: [AccountType.LEARNER,AccountType.MENTOR]')

    def test_render_access__no_access(self):

        conf = deepcopy(CONF)
        conf['access'] = {
            'access_list': [],
            'is_private': False,
        }

        command = Command('BULK_READ_TASKS', conf)

        assert command.render_access('AccountType') == (
            'BULK_READ_TASKS: null')

    #
    # get_bulk_read_field
    #
    def test_get_bulk_read_field(self):

        conf = deepcopy(CONF)
        conf['schemas'] = {
            'output': {
                'schema': {
                    'type': 'object',
                    'properties': {
                        'tasks': {
                            'type': 'array',
                        }
                    }
                },
                'uri': 'hi/there',
            },
        }

        command = Command('BULK_READ_TASKS', conf)

        assert command.get_bulk_read_field() == 'tasks'

    def test_get_bulk_read_field__not_bulk_read_name(self):

        command = Command('READ_TASK', CONF)

        assert command.get_bulk_read_field() is None

    def test_get_bulk_read_field__more_than_one_field(self):

        conf = deepcopy(CONF)
        conf['schemas'] = {
            'output': {
                'schema': {
                    'type': 'object',
                    'properties': {
                        'tasks': {
                            'type': 'array',
                        },
                        'people': {
                            'type': 'array',
                        }
                    }
                },
                'uri': 'hi/there',
            },
        }

        command = Command('BULK_READ_TASKS', conf)

        assert command.get_bulk_read_field() is None

    def test_get_bulk_read_field__empty_schema(self):

        conf = deepcopy(CONF)
        conf['schemas'] = {
            'output': {
                'schema': {},
                'uri': 'hi/there',
            },
        }

        command = Command('BULK_READ_TASKS', conf)

        assert command.get_bulk_read_field() is None




import pytest

from lily.docs.renderers.angular.utils import to_camelcase


@pytest.mark.parametrize(
    'name, expected', [
        # -- empty to empty
        ('', ''),

        # -- single letter capital to same
        ('A', 'A'),

        # -- camel to camel
        ('Abc', 'Abc'),

        # -- camel to camel
        ('Abc', 'Abc'),

        # -- all UPPER to camel
        ('ABCD', 'Abcd'),

        # -- all lower to camel
        ('abcd', 'Abcd'),

        # -- underscore case
        ('ab_cde_f', 'AbCdeF'),

        # -- mixed case
        ('ab_CDe_f', 'AbCDeF'),

        # -- camel case stays camel case
        ('AgeChoice', 'AgeChoice'),

        # -- all upper is lower first
        ('AGE_CHOICE', 'AgeChoice'),
    ])
def test_to_camelcase(name, expected):

    assert to_camelcase(name) == expected


@pytest.mark.parametrize(
    'name, expected', [
        # -- empty to empty
        ('', ''),

        # -- single letter capital to same
        ('A', 'a'),

        # -- all lower to camel
        ('ABcd', 'aBcd'),

        # -- underscore case
        ('ab_cde_f', 'abCdeF'),

        # -- mixed case
        ('ab_CDe_f', 'abCDeF'),
    ])
def test_to_camelcase__first_lower(name, expected):

    assert to_camelcase(name, first_lower=True) == expected




import re


HASHTAG_ESCAPE_SEQUENCE = 'HASH58437HASH'
"""
Sequence of characters which will be used to replace # character appearing
in front of any word transforming it into hashtag.

Default behaviour of data base strips away the hash character and stems
the hashtag word therefore making usage of hashtags as filters impossible. To
prevent this default behaviour the sequence ``HASHTAG_ESCAPE_SEQUENCE`` is
replaced in both searched text and query.

"""


HASHTAG_PATTERN = re.compile(
    r'(?P<prefix>'
    r'^|'  # either it appears at the beginning of string
    r'\s+|'  # or is preceded by space of any kind
    r'~|'  # or is preceded by negation
    r'['
    r':;,.!?。'  # or is preceded by punctuation
    r'’\[\]\(\)'  # or is preceded by brackets
    r'])(?P<hash>#)(?P<text>\w+)',
    flags=re.I | re.U)




from unittest.mock import call
from copy import deepcopy
import json

from django.test import TestCase
from django.urls import reverse
import pytest

from lily.base.test import Client
from entrypoint.commands import (
    EntryPointCommands,
    CommandSerializer,
)
from tests.factory import EntityFactory


ef = EntityFactory()


class MockConfig:

    def __init__(self, **kwargs):
        self.version = '2.5.6'
        self.name = 'test'

    @classmethod
    def get_lily_path(cls):
        return cls._lily_path


class EntryPointCommandsTestCase(TestCase):

    uri = reverse('entrypoint:entrypoint')

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.app = Client()
        self.auth_headers = {
            'HTTP_X_CS_ACCOUNT_TYPE': 'ADMIN',
            'HTTP_X_CS_USER_ID': 190,
        }

        self.lily_dir = self.tmpdir.mkdir('.lily')
        self.commands_dir = self.lily_dir.mkdir('commands')

        MockConfig._lily_path = str(self.lily_dir)
        self.mocker.patch(
            'entrypoint.commands.Config', MockConfig)

    def test_get(self):

        c = ef.command()
        self.mocker.patch.object(
            EntryPointCommands,
            'get_commands'
        ).return_value = {
            '@enums': [],
            'UPDATE_HELLO': CommandSerializer(deepcopy(c)).data,
        }

        response = self.app.get(self.uri, **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'UPDATE_HELLO': CommandSerializer(c).data,
            },
        }

    def test_get__with_versions(self):

        c = ef.command()
        self.mocker.patch.object(
            EntryPointCommands,
            'get_commands'
        ).return_value = {
            '@enums': [],
            'UPDATE_HELLO': CommandSerializer(deepcopy(c)).data,
        }
        self.commands_dir.join('2.5.6.json').write('..')
        self.commands_dir.join('2.14.5.json').write('..')
        self.commands_dir.join('2.120.0.json').write('..')
        self.commands_dir.join('1.0.0.json').write('..')

        response = self.app.get(self.uri, **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': ['2.120.0', '2.14.5', '2.5.6', '1.0.0'],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'UPDATE_HELLO': CommandSerializer(c).data,
            },
        }

    def test_get__filter_by_commands_query(self):

        c0, c1, c2 = ef.command(), ef.command(), ef.command()
        self.mocker.patch.object(
            EntryPointCommands,
            'get_commands'
        ).side_effect = [
            {
                '@enums': [],
                'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
                'READ_PAYMENTS': CommandSerializer(deepcopy(c1)).data,
                'FIND_TOOL': CommandSerializer(deepcopy(c2)).data,
            },
            {
                '@enums': [],
                'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
                'READ_PAYMENTS': CommandSerializer(deepcopy(c1)).data,
                'FIND_TOOL': CommandSerializer(deepcopy(c2)).data,
            },
        ]

        # -- filter two commands
        response = self.app.get(
            self.uri,
            data={
                'commands': ['UPDATE_HELLO', 'FIND_TOOL']
            },
            **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'UPDATE_HELLO': CommandSerializer(c0).data,
                'FIND_TOOL': CommandSerializer(c2).data,
            },
        }

        # -- filter single command - triangulation
        response = self.app.get(
            self.uri,
            data={
                'commands': ['READ_PAYMENTS']
            },
            **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'READ_PAYMENTS': CommandSerializer(c1).data,
            },
        }

    def test_get__filter_by_is_private(self):

        c0 = ef.command(is_private=True)
        c1 = ef.command(is_private=False)
        c2 = ef.command(is_private=True)
        self.mocker.patch.object(
            EntryPointCommands,
            'get_commands'
        ).side_effect = [
            {
                '@enums': [],
                'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
                'CREATE_HELLO': CommandSerializer(deepcopy(c1)).data,
                'DELETE_HELLO': CommandSerializer(deepcopy(c2)).data,
            },
            {
                '@enums': [],
                'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
                'CREATE_HELLO': CommandSerializer(deepcopy(c1)).data,
                'DELETE_HELLO': CommandSerializer(deepcopy(c2)).data,
            },
        ]

        # -- show only private commands
        response = self.app.get(
            self.uri,
            data={'is_private': True},
            **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'UPDATE_HELLO': CommandSerializer(c0).data,
                'DELETE_HELLO': CommandSerializer(c2).data,
            },
        }

        # -- show only public commands
        response = self.app.get(
            self.uri,
            data={'is_private': False},
            **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'CREATE_HELLO': CommandSerializer(c1).data,
            },
        }

    def test_get__filter_by_domain_id(self):

        c0 = ef.command(domain_id='cards')
        c1 = ef.command(domain_id='paths')
        c2 = ef.command(domain_id='paths')
        self.mocker.patch.object(
            EntryPointCommands,
            'get_commands'
        ).side_effect = [
            {
                '@enums': [{'A': 'C'}],
                'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
                'CREATE_HELLO': CommandSerializer(deepcopy(c1)).data,
                'DELETE_HELLO': CommandSerializer(deepcopy(c2)).data,
            },
            {
                '@enums': [{'A': 'X'}],
                'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
                'CREATE_HELLO': CommandSerializer(deepcopy(c1)).data,
                'DELETE_HELLO': CommandSerializer(deepcopy(c2)).data,
            },
        ]

        # -- show CARDS domain commands
        response = self.app.get(
            self.uri,
            data={'domain_id': 'cards'},
            **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [{'A': 'C'}],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'UPDATE_HELLO': CommandSerializer(c0).data,
            },
        }

        # -- show PATHS domain commands
        response = self.app.get(
            self.uri,
            data={'domain_id': 'PATHS'},
            **self.auth_headers)

        assert response.status_code == 200
        assert response.json() == {
            '@event': 'ENTRY_POINT_READ',
            '@type': 'entrypoint',
            '@enums': [{'A': 'X'}],
            'name': 'test',
            'version_info': {
                '@type': 'version_info',
                'available': [],
                'deployed': '2.5.6',
                'displayed': '2.5.6',
            },
            'commands': {
                'CREATE_HELLO': CommandSerializer(c1).data,
                'DELETE_HELLO': CommandSerializer(c2).data,
            },
        }

    def test_get__version(self):

        c = ef.command()
        get_commands = self.mocker.patch.object(
            EntryPointCommands, 'get_commands')
        get_commands.return_value = {
            '@enums': [],
            'UPDATE_HELLO': CommandSerializer(deepcopy(c)).data,
        }

        response = self.app.get(
            self.uri,
            data={'version': '2.1.3'},
            **self.auth_headers)

        assert response.status_code == 200
        version_info = response.json()['version_info']
        assert version_info == {
            '@type': 'version_info',
            'available': [],
            'deployed': '2.5.6',
            'displayed': '2.1.3',
        }
        assert get_commands.call_args_list == [call('2.1.3')]

    #
    # GET_COMMANDS
    #
    def test_get_commands(self):

        c0 = ef.command(domain_id='cards')
        c1 = ef.command(domain_id='paths')

        commands = {
            'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
            'CREATE_HELLO': CommandSerializer(deepcopy(c1)).data,
        }
        self.commands_dir.join('2.5.6.json').write(json.dumps(commands))

        assert EntryPointCommands().get_commands() == commands

    def test_get_commands__version(self):

        c0 = ef.command(domain_id='cards')
        c1 = ef.command(domain_id='paths')

        commands0 = {  # noqa
            'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
            'CREATE_HELLO': CommandSerializer(deepcopy(c1)).data,
        }
        self.commands_dir.join('2.5.6.json').write(json.dumps(commands0))
        commands1 = {
            'UPDATE_HELLO': CommandSerializer(deepcopy(c0)).data,
        }
        self.commands_dir.join('2.0.0.json').write(json.dumps(commands1))

        assert EntryPointCommands().get_commands('2.0.0') == commands1




import os  # pragma: no cover

from django.core.wsgi import get_wsgi_application  # pragma: no cover

os.environ.setdefault(  # pragma: no cover
    "DJANGO_SETTINGS_MODULE", "conf.settings")

application = get_wsgi_application()  # pragma: no cover




from collections import OrderedDict
from unittest.mock import call, Mock
import json
import os
import tempfile

from django.test import TestCase
import pytest

from lily.base.events import EventFactory
from lily.base.utils import normalize_indentation
from lily.docs.renderers.angular.domain import Domain
from lily.docs.renderers.angular.interface import Enum
from lily.docs.renderers.angular.renderer import AngularClientRenderer
from tests import remove_white_chars


class AngularClientRendererTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):

        self.base_dir = self.tmpdir.mkdir('client')
        self.src_dir = (
            self.base_dir
            .mkdir('projects')
            .mkdir('client')
            .mkdir('src'))
        self.domains_dir = self.src_dir.mkdir('domains')
        self.services_dir = self.src_dir.mkdir('services')

        self.mocker.patch.object(
            tempfile, 'mkdtemp').return_value = str(self.base_dir)
        self.renderer = AngularClientRenderer('origin', 'prefix')
        self.current_cwd = os.getcwd()
        self.renderer.repo.cd_to_repo()

    def tearDown(self):
        os.chdir(self.current_cwd)

    #
    # RENDER
    #
    def test_render(self):

        class MockConfig:

            @property
            def version(self):
                return '0.9.1'

            @classmethod
            def get_lily_path(cls):
                return str(self.tmpdir)

        commands_dir = self.tmpdir.mkdir('commands')
        commands_dir.join('0.9.1.json').write(json.dumps({'some': 'commands'}))
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Config', MockConfig)

        repo = Mock()
        self.mocker.patch.object(self.renderer, 'repo', repo)

        http_repo = Mock()
        self.mocker.patch.object(self.renderer, 'http_repo', http_repo)

        template_repo = Mock()
        self.mocker.patch.object(self.renderer, 'template_repo', template_repo)

        group_commands_by_domain = self.mocker.patch.object(
            self.renderer, 'group_commands_by_domain')
        group_commands_by_domain.return_value = OrderedDict([
            (Domain('cards', ''), {'X': 'X', 'B': 'B'}),
            (Domain('paths', ''), {'Z': 'Z', 'A': 'A'}),
        ])

        render_client_module_ts = self.mocker.patch.object(
            self.renderer, 'render_client_module_ts')

        render_api_ts = self.mocker.patch.object(
            self.renderer, 'render_api_ts')

        render_api_index_ts = self.mocker.patch.object(
            self.renderer, 'render_api_index_ts')

        render_domain = self.mocker.patch.object(
            self.renderer, 'render_domain')

        render_access_ts = self.mocker.patch.object(
            self.renderer, 'render_access_ts')

        render_enums_ts = self.mocker.patch.object(
            self.renderer, 'render_enums_ts')

        self.renderer.render()

        assert group_commands_by_domain.call_count == 1
        assert render_client_module_ts.call_args_list == [
            call(group_commands_by_domain.return_value)]
        assert render_api_index_ts.call_args_list == [
            call(group_commands_by_domain.return_value)]
        assert render_api_ts.call_args_list == [
            call(group_commands_by_domain.return_value)]
        assert render_domain.call_args_list == [
            call(Domain('cards', ''), ['B', 'X']),
            call(Domain('paths', ''), ['A', 'Z']),
        ]

        # -- shared
        assert self.renderer.repo.create_dir.call_args_list == [
            call('./projects/client/src/shared'),
        ]
        assert render_access_ts.call_args_list == [
            call(
                './projects/client/src/shared',
                ['B', 'X', 'A', 'Z']
            )]
        assert render_enums_ts.call_args_list == [
            call(
                './projects/client/src/shared',
                ['B', 'X', 'A', 'Z']
            )]

        assert template_repo.clone.call_count == 1

        assert repo.clone.call_count == 1
        assert repo.install.call_count == 1
        assert repo.upgrade_version.call_count == 1
        assert repo.build.call_count == 1
        assert repo.add_all.call_count == 1
        assert repo.commit.call_count == 1
        assert repo.push.call_count == 1
        assert http_repo.cd_to_repo.call_count == 1
        assert http_repo.clone.call_count == 1
        assert http_repo.install.call_count == 1
        assert http_repo.build.call_count == 1
        assert http_repo.link.call_count == 1

    def test_render__only_build(self):

        class MockConfig:

            @property
            def version(self):
                return '0.9.1'

            @classmethod
            def get_lily_path(cls):
                return str(self.tmpdir)

        commands_dir = self.tmpdir.mkdir('commands')
        commands_dir.join('0.9.1.json').write(json.dumps({'some': 'commands'}))
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Config', MockConfig)

        repo = Mock()
        self.mocker.patch.object(self.renderer, 'repo', repo)

        http_repo = Mock()
        self.mocker.patch.object(self.renderer, 'http_repo', http_repo)

        template_repo = Mock()
        self.mocker.patch.object(self.renderer, 'template_repo', template_repo)

        group_commands_by_domain = self.mocker.patch.object(
            self.renderer, 'group_commands_by_domain')
        group_commands_by_domain.return_value = OrderedDict([
            (Domain('cards', ''), {'X': 'X', 'B': 'B'}),
            (Domain('paths', ''), {'Z': 'Z', 'A': 'A'}),
        ])

        render_client_module_ts = self.mocker.patch.object(
            self.renderer, 'render_client_module_ts')

        render_api_ts = self.mocker.patch.object(
            self.renderer, 'render_api_ts')

        render_api_index_ts = self.mocker.patch.object(
            self.renderer, 'render_api_index_ts')

        render_domain = self.mocker.patch.object(
            self.renderer, 'render_domain')

        render_access_ts = self.mocker.patch.object(
            self.renderer, 'render_access_ts')

        render_enums_ts = self.mocker.patch.object(
            self.renderer, 'render_enums_ts')

        self.renderer.render(only_build=True)

        assert group_commands_by_domain.call_count == 1
        assert render_client_module_ts.call_args_list == [
            call(group_commands_by_domain.return_value)]
        assert render_api_index_ts.call_args_list == [
            call(group_commands_by_domain.return_value)]
        assert render_api_ts.call_args_list == [
            call(group_commands_by_domain.return_value)]
        assert render_domain.call_args_list == [
            call(Domain('cards', ''), ['B', 'X']),
            call(Domain('paths', ''), ['A', 'Z']),
        ]

        # -- shared
        assert self.renderer.repo.create_dir.call_args_list == [
            call('./projects/client/src/shared'),
        ]
        assert render_access_ts.call_args_list == [
            call(
                './projects/client/src/shared',
                ['B', 'X', 'A', 'Z']
            )]
        assert render_enums_ts.call_args_list == [
            call(
                './projects/client/src/shared',
                ['B', 'X', 'A', 'Z']
            )]

        assert template_repo.clone.call_count == 1

        assert repo.clone.call_count == 1
        assert repo.install.call_count == 1
        assert repo.build.call_count == 1
        assert repo.upgrade_version.call_count == 0
        assert repo.add_all.call_count == 0
        assert repo.commit.call_count == 0
        assert repo.push.call_count == 0
        assert http_repo.cd_to_repo.call_count == 1
        assert http_repo.clone.call_count == 1
        assert http_repo.install.call_count == 1
        assert http_repo.build.call_count == 1
        assert http_repo.link.call_count == 1

    #
    # GROUP_COMMANDS_BY_DOMAIN
    #
    def test_group_commands_by_domain(self):

        command0, command1, command2 = (
            Mock(domain_id='cards', domain_name='...', is_private=False),
            Mock(domain_id='paths', domain_name='...', is_private=False),
            Mock(domain_id='cards', domain_name='...', is_private=False),
        )
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Command'
        ).side_effect = [command0, command1, command2]
        self.mocker.patch.object(
            AngularClientRenderer,
            'get_commands'
        ).return_value = OrderedDict([
            ('CREATE_CARD', {'create_card': 'conf'}),
            ('REMOVE_PATH', {'remove_path': 'conf'}),
            ('REMOVE_CARD', {'remove_card': 'conf'}),
        ])

        commands_by_domain = self.renderer.group_commands_by_domain()

        assert commands_by_domain == {
            Domain('cards', ''): {
                'CREATE_CARD': command0,
                'REMOVE_CARD': command2,
            },
            Domain('paths', ''): {
                'REMOVE_PATH': command1,
            }
        }

    def test_group_commands_by_domain__exclude_domains(self):

        command0, command1, command2 = (
            Mock(domain_id='cards', domain_name='...', is_private=False),
            Mock(domain_id='paths', domain_name='...', is_private=False),
            Mock(domain_id='cards', domain_name='...', is_private=False),
        )
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Command'
        ).side_effect = [command0, command1, command2]
        self.mocker.patch.object(
            AngularClientRenderer,
            'get_commands'
        ).return_value = OrderedDict([
            ('CREATE_CARD', {'create_card': 'conf'}),
            ('REMOVE_PATH', {'remove_path': 'conf'}),
            ('REMOVE_CARD', {'remove_card': 'conf'}),
        ])

        commands_by_domain = self.renderer.group_commands_by_domain(
            exclude_domains=('PATHS',))

        assert commands_by_domain == {
            Domain('cards', ''): {
                'CREATE_CARD': command0,
                'REMOVE_CARD': command2,
            },
        }

    def test_group_commands_by_domain__include_domains(self):

        command0, command1, command2 = (
            Mock(domain_id='hashtags', domain_name='...', is_private=False),
            Mock(domain_id='paths', domain_name='...', is_private=False),
            Mock(domain_id='cards', domain_name='...', is_private=False),
        )
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Command'
        ).side_effect = [command0, command1, command2]
        self.mocker.patch.object(
            AngularClientRenderer,
            'get_commands'
        ).return_value = OrderedDict([
            ('CREATE_HASHTAG', {'create_hashtag': 'conf'}),
            ('REMOVE_PATH', {'remove_path': 'conf'}),
            ('REMOVE_CARD', {'remove_card': 'conf'}),
        ])

        commands_by_domain = self.renderer.group_commands_by_domain(
            include_domains=('PATHS', 'hashtags'))

        assert commands_by_domain == {
            Domain('hashtags', ''): {
                'CREATE_HASHTAG': command0,
            },
            Domain('paths', ''): {
                'REMOVE_PATH': command1,
            }
        }

    def test_group_commands_by_domain__is_private(self):

        command0, command1, command2 = (
            Mock(domain_id='cards', domain_name='...', is_private=False),
            Mock(domain_id='paths', domain_name='...', is_private=True),
            Mock(domain_id='cards', domain_name='...', is_private=True),
        )
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Command'
        ).side_effect = [command0, command1, command2]
        self.mocker.patch.object(
            AngularClientRenderer,
            'get_commands'
        ).return_value = OrderedDict([
            ('CREATE_CARD', {'create_card': 'conf'}),
            ('REMOVE_PATH', {'remove_path': 'conf'}),
            ('REMOVE_CARD', {'remove_card': 'conf'}),
        ])

        commands_by_domain = self.renderer.group_commands_by_domain()

        assert commands_by_domain == {
            Domain('cards', ''): {
                'CREATE_CARD': command0,
            }
        }

    #
    # GET_COMMANDS
    #
    def test_get_commands(self):

        class MockConfig:

            @property
            def version(self):
                return '0.9.1'

            @classmethod
            def get_lily_path(cls):
                return str(self.tmpdir)

        commands_dir = self.tmpdir.mkdir('commands')
        commands_dir.join('0.9.1.json').write(json.dumps({'some': 'commands'}))
        self.mocker.patch(
            'lily.docs.renderers.angular.renderer.Config', MockConfig)

        assert self.renderer.get_commands() == {'some': 'commands'}

    #
    # RENDER_DOMAIN
    #
    def test_render_domain(self):

        render_index_ts = self.mocker.patch.object(
            AngularClientRenderer, 'render_index_ts')
        render_models_ts = self.mocker.patch.object(
            AngularClientRenderer, 'render_models_ts')
        render_domain_ts = self.mocker.patch.object(
            AngularClientRenderer, 'render_domain_ts')
        render_examples_ts = self.mocker.patch.object(
            AngularClientRenderer, 'render_examples_ts')

        domain = Domain('cards', 'Cards Management')

        domains_path = str(self.domains_dir)

        assert os.listdir(domains_path) == []

        self.renderer.render_domain(domain, [])

        assert os.listdir(domains_path) == ['cards']
        assert render_index_ts.call_args_list == [call(domain)]
        assert render_models_ts.call_args_list == [call(domain, [])]
        assert render_domain_ts.call_args_list == [call(domain, [])]
        assert render_examples_ts.call_args_list == [call(domain, [])]

    #
    # RENDER_INDEX_TS
    #
    def test_render_index_ts(self):

        paths_path = str(self.domains_dir.mkdir('paths'))

        assert os.listdir(paths_path) == []

        self.renderer.render_index_ts(Domain('paths', 'Path Management'))

        assert os.listdir(paths_path) == ['index.ts']
        with open(os.path.join(paths_path, 'index.ts'), 'r') as f:
            assert f.read() == normalize_indentation('''
                export * from './paths.domain';
                export * from './paths.models';
            ''', 0)

    #
    # RENDER_MODELS_TS
    #
    def test_render_models_ts__without_enums(self):

        paths_path = str(self.domains_dir.mkdir('paths'))
        domain = Domain('paths', 'Path Management')
        commands = [
            Mock(
                request_query=Mock(
                    render=Mock(return_value=['request 1', []])),
                request_body=Mock(
                    render=Mock(return_value=['', []])),
                response=Mock(
                    render=Mock(return_value=['response 1', []])),
            ),
            Mock(
                request_query=Mock(
                    render=Mock(return_value=['request 2', []])),
                request_body=Mock(
                    render=Mock(return_value=['', []])),
                response=Mock(
                    render=Mock(return_value=['response 2', []])),
            ),
        ]

        assert os.listdir(paths_path) == []

        self.renderer.render_models_ts(domain, commands)

        assert os.listdir(paths_path) == ['paths.models.ts']
        with open(os.path.join(paths_path, 'paths.models.ts'), 'r') as f:
            assert f.read() == normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */



                /**
                 * Path Management Domain Models
                 */

                request 1

                response 1

                request 2

                response 2
            ''', 0)

    def test_render_models_ts__with_enums(self):

        paths_path = str(self.domains_dir.mkdir('paths'))
        domain = Domain('paths', 'Path Management')
        commands = [
            Mock(
                request_query=Mock(
                    render=Mock(return_value=[
                        'request 1', [Enum('Type', ['A'])]])),
                request_body=Mock(
                    render=Mock(return_value=['', []])),
                response=Mock(
                    render=Mock(return_value=[
                        'response 1', [Enum('Jack', []), Enum('Type', [])]])),
            ),
        ]

        assert os.listdir(paths_path) == []

        self.renderer.render_models_ts(domain, commands)

        assert os.listdir(paths_path) == ['paths.models.ts']
        with open(os.path.join(paths_path, 'paths.models.ts'), 'r') as f:
            assert f.read() == normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */

                import { Jack, Type } from '../../shared/enums';

                /**
                 * Path Management Domain Models
                 */

                request 1

                response 1
            ''', 0)

    #
    # RENDER_DOMAIN_TS
    #
    def test_render_domain_ts(self):

        paths_path = str(self.domains_dir.mkdir('paths'))
        domain = Domain('paths', 'Path Management')
        commands = [
            Mock(render=Mock(return_value='command 1')),
            Mock(render=Mock(return_value='command 2')),
        ]

        assert os.listdir(paths_path) == []

        self.renderer.render_domain_ts(domain, commands)

        assert os.listdir(paths_path) == ['paths.domain.ts']
        with open(os.path.join(paths_path, 'paths.domain.ts'), 'r') as f:
            assert f.read() == normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */

                /**
                 * Path Management Domain
                 */
                import { Injectable } from '@angular/core';
                import { filter } from 'rxjs/operators';
                import { Observable } from 'rxjs';
                import * as _ from 'underscore';

                import { HttpService } from '../../http.service';

                import * as X from './paths.models';

                @Injectable()
                export class PathsDomain {
                    constructor(private client: HttpService) {}

                    command 1

                    command 2

                }
            ''', 0)  # noqa

    #
    # RENDER_EXAMPLES_TS
    #
    def test_render_examples_ts(self):

        paths_path = str(self.domains_dir.mkdir('paths'))
        domain = Domain('paths', 'Path Management')
        commands = [
            Mock(render_examples=Mock(return_value='command 1')),
            Mock(render_examples=Mock(return_value='command 2')),
        ]

        assert os.listdir(paths_path) == []

        self.renderer.render_examples_ts(domain, commands)

        assert os.listdir(paths_path) == ['paths.examples.ts']
        with open(os.path.join(paths_path, 'paths.examples.ts'), 'r') as f:
            assert f.read() == normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */

                /**
                 * Path Management Domain Examples
                 */

                command 1

                command 2
            ''', 0)  # noqa

    #
    # RENDER_ACCESS_TS
    #
    def test_render_access_ts(self):

        shared_path = str(self.src_dir.mkdir('shared'))
        commands = [
            Mock(render_access=Mock(
                return_value=(
                    'BULK_READ_WHAT: [AccountType.LEARNER, AccountType.MENTOR]'
                ))),
            Mock(render_access=Mock(return_value='BULK_DELETE_THIS: null')),
            Mock(
                render_access=Mock(
                    return_value='CREATE: [AccountType.ADMIN]')),
        ]

        assert os.listdir(shared_path) == []

        self.renderer.render_access_ts(shared_path, commands)

        assert os.listdir(shared_path) == ['access.ts']
        with open(os.path.join(shared_path, 'access.ts'), 'r') as f:
            result = f.read()
            assert result == normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */

                import { AccountType } from './enums';

                export const Access = {
                  BULK_READ_WHAT: [AccountType.LEARNER, AccountType.MENTOR],
                  BULK_DELETE_THIS: null,
                  CREATE: [AccountType.ADMIN],
                }
            ''', 0)  # noqa

    #
    # RENDER_ENUMS_TS
    #
    def test_render_enums_ts__no_enums(self):

        self.mocker.patch.object(
            self.renderer,
            'collect_unique_enums'
        ).return_value = []
        shared_path = str(self.src_dir.mkdir('shared'))

        self.renderer.render_enums_ts(shared_path, Mock())

        assert self.src_dir.join('shared/enums.ts').read() == ''

    def test_render_enums_ts(self):

        self.mocker.patch.object(
            self.renderer,
            'collect_unique_enums'
        ).return_value = [
            Enum('age', [11, 33]),
            Enum('category', ['AA', 'YY', 'XX']),
            Enum('name', ['Jack', 'Alice', 'Joe']),
        ]
        shared_path = str(self.src_dir.mkdir('shared'))

        self.renderer.render_enums_ts(shared_path, Mock())

        assert os.listdir(self.src_dir) == ['domains', 'services', 'shared']
        assert (
            self.src_dir.join('shared/enums.ts').read() ==
            normalize_indentation('''
                export enum Age {
                    VALUE_11 = 11,
                    VALUE_33 = 33,
                }

                export enum Category {
                    AA = 'AA',
                    XX = 'XX',
                    YY = 'YY',
                }

                export enum Name {
                    ALICE = 'Alice',
                    JACK = 'Jack',
                    JOE = 'Joe',
                }
            ''', 0))  # noqa

    #
    # COLLECT_UNIQUE_ENUMS
    #
    def test_collect_unique_enums__all_unique(self):

        e0 = Enum('age', [11, 33])
        e1 = Enum('name', ['Jack', 'Alice', 'Joe'])
        e2 = Enum('category', ['AA', 'YY', 'XX'])

        assert self.renderer.collect_unique_enums([
            Mock(
                request_query=Mock(enums=[e0, e1]),
                request_body=Mock(enums=[]),
                response=Mock(enums=[])),
            Mock(
                request_query=Mock(enums=[]),
                request_body=Mock(enums=[e2]),
                response=Mock(enums=[])),
        ]) == [e0, e2, e1]

    def test_collect_unique_enums__with_duplicates(self):

        e0 = Enum('age', [11, 33])
        e1 = Enum('age', [11, 33])
        e2 = Enum('name', ['Jack', 'Alice', 'Joe'])
        e3 = Enum('name', ['Jack', 'Alice', 'Joe'])
        e4 = Enum('category', ['AA', 'YY', 'XX'])

        assert self.renderer.collect_unique_enums([
            Mock(
                request_query=Mock(enums=[e0]),
                request_body=Mock(enums=[]),
                response=Mock(enums=[])),
            Mock(
                request_query=Mock(enums=[]),
                request_body=Mock(enums=[e1, e2]),
                response=Mock(enums=[])),
            Mock(
                request_query=Mock(enums=[]),
                request_body=Mock(enums=[]),
                response=Mock(enums=[e3, e4])),
        ]) == [e0, e4, e2]

    def test_collect_unique_enums__inconsistent_duplicates_detected(self):

        e0 = Enum('age', [11, 33])
        e1 = Enum('age', [11, 22, 33])
        e2 = Enum('name', ['Jack', 'Alice', 'Joe'])

        with pytest.raises(EventFactory.ServerError) as e:
            self.renderer.collect_unique_enums([
                Mock(
                    request_query=Mock(enums=[e0]),
                    request_body=Mock(enums=[]),
                    response=Mock(enums=[])),
                Mock(
                    request_query=Mock(enums=[]),
                    request_body=Mock(enums=[e1, e2]),
                    response=Mock(enums=[])),
            ])

        assert e.value.data == {
            '@event': 'INCONSISTENT_ENUMS_DETECTED',
            '@type': 'error',
            'inconsistent_enums': [
                {
                    'enum.0': {
                        'name': 'Age',
                        'values': [33, 11],
                    },
                    'enum.1': {
                        'name': 'Age',
                        'values': [33, 11, 22],
                    },
                },
            ],
        }

    #
    # RENDER_API_INDEX_TS
    #
    def test_render_api_index_ts(self):

        commands_by_domain = {
            Domain('cards', 'Cards Management'): {},
            Domain('recall', 'Recall Management'): {},
            Domain('paths', 'Path Management'): {},
        }

        self.renderer.render_api_index_ts(commands_by_domain)

        assert os.listdir(self.domains_dir) == ['index.ts']
        assert (
            self.domains_dir.join('index.ts').read() ==
            normalize_indentation('''
                export * from './cards/index';
                export * from './paths/index';
                export * from './recall/index';
            ''', 0))  # noqa

    #
    # RENDER_API_TS
    #
    def test_render_api_ts(self):

        command = lambda x: Mock(render_facade=Mock(return_value=x))  # noqa
        commands_by_domain = {
            Domain('cards', 'Cards Management'): {
                'READ_CARDS': command('    <READ_CARDS>'),
                'DELETE_CARD': command('    <DELETE_CARD>'),
            },
            Domain('paths', 'Path Management'): {
                'CREATE_PATH': command('    <CREATE_PATH>'),
            },
        }
        services_path = str(self.services_dir)

        self.renderer.render_api_ts(commands_by_domain)

        assert os.listdir(services_path) == ['api.service.ts']
        with open(os.path.join(services_path, 'api.service.ts'), 'r') as f:
            assert remove_white_chars(f.read()) == remove_white_chars(normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */

                /**
                 * Facade API Service for all domains
                 */
                import { Injectable, Injector } from '@angular/core';
                import { Observable } from 'rxjs';

                import * as X from '../domains/index';

                @Injectable()
                export class APIService {

                    constructor(private injector: Injector) {}

                    /**
                     * Cards Management domain
                     */
                    private _cardsDomain: X.CardsDomain;

                    public get cardsDomain(): X.CardsDomain {
                        if (!this._cardsDomain) {
                            this._cardsDomain = this.injector.get(X.CardsDomain);
                        }

                        return this._cardsDomain;
                    }

                    <DELETE_CARD>

                    <READ_CARDS>

                    /**
                     * Path Management domain
                     */
                    private _pathsDomain: X.PathsDomain;

                    public get pathsDomain(): X.PathsDomain {
                        if (!this._pathsDomain) {
                            this._pathsDomain = this.injector.get(X.PathsDomain);
                        }

                        return this._pathsDomain;
                    }

                    <CREATE_PATH>

                }
            ''', 0))  # noqa

    #
    # RENDER_CLIENT_MODULE_TS
    #
    def test_render_client_module_ts(self):

        commands_by_domain = {
            Domain('cards', 'Cards Management'): {},
            Domain('paths', 'Path Management'): {},
            Domain('recall', 'Recall Management'): {},
        }
        src_path = str(self.src_dir)

        self.renderer.render_client_module_ts(commands_by_domain)

        assert sorted(os.listdir(src_path)) == [
            'client.module.ts',
            'domains',
            'services',
        ]
        with open(
                os.path.join(src_path, 'client.module.ts'), 'r') as f:
            assert f.read() == normalize_indentation('''
                /**
                  * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
                  * OVERWRITTEN
                  */

                import { NgModule, ModuleWithProviders } from '@angular/core';
                import { HttpClientModule } from '@angular/common/http';

                /** Domains */
                import { CardsDomain } from './domains/cards/index';
                import { PathsDomain } from './domains/paths/index';
                import { RecallDomain } from './domains/recall/index';

                /** Services */
                import { APIService } from './services/api.service';
                import { HttpService, Config } from './http.service';

                @NgModule({
                    imports: [HttpClientModule],
                    providers: [
                        // Domains
                        CardsDomain,
                        PathsDomain,
                        RecallDomain,

                        // Facade
                        APIService,
                        HttpService,
                    ]
                })
                export class ClientModule {
                  static forRoot(config: Config): ModuleWithProviders<ClientModule> {

                    return {

                      ngModule: ClientModule,

                      providers: [{ provide: 'COSPHERE_CLIENT_CONFIG_TOKEN', useValue: config }]

                    };
                  }
                }
            ''', 0)  # noqa




import re
from copy import deepcopy

from django.contrib.postgres.search import (
    SearchVector,
    SearchVectorCombinable,
    CombinedExpression,
)
from django.db import connection
from trans import trans
import hunspell

from .latex.transformer import transform as transform_latex
from .constants import HASHTAG_ESCAPE_SEQUENCE, HASHTAG_PATTERN
from .stopwords import stopwords_filter


class TextVector:

    hobj_pl = None

    def __init__(self):
        if not TextVector.hobj_pl:
            try:
                TextVector.hobj_pl = hunspell.HunSpell(
                    '/usr/share/hunspell/pl_PL.dic',
                    '/usr/share/hunspell/pl_PL.aff')
            except Exception:
                TextVector.hobj_pl = None

    def parse_to_tsvector(self, conf, text, weight=None):

        stems = self.parse(conf, text, weight)
        return self.to_tsvector(stems)

    def parse_to_list(self, conf, text, weight=None):

        stems = self.parse(conf, text, weight)
        return sorted(stems.keys())

    def parse(self, conf, text, weight=None):
        if text is None or not text.strip():
            return {}

        # -- transformers
        text = text.strip()
        text = self.transform_special_characters(text)
        text = self.transform_hashtags(text)
        text = self.transform_latex(conf, text)

        # -- tokens
        tokens = self.tokenize(conf, text)

        # -- stems
        stems = {}
        stems = self.augument_with_stems(conf, text, weight, stems, tokens)
        stems = self.augument_with_unaccents(conf, weight, stems, tokens)

        return stems

    #
    # SPECIAL CHARACTERS
    #
    def transform_special_characters(self, text):
        weird = [
            'ą', 'Ą', 'a\\u0328', 'A\\u0328',
            'ę', 'Ę', 'e\\u0328', 'E\\u0328',
            'ó', 'Ó', 'o\\u0301', 'O\\u0301',
            'ć', 'Ć', 'c\\u0301', 'C\\u0301',
            'ś', 'Ś', 's\\u0301', 'S\\u0301',
            'ż', 'Ż', 'z\\u0307', 'Z\\u0307',
            'ź', 'Ź', 'z\\u0301', 'Z\\u0301',
            'ń', 'Ń', 'n\\u0301', 'N\\u0301',
        ]
        normal = [
            'ą', 'Ą', '\u0105', '\u0104',
            'ę', 'Ę', '\u0119', '\u0118',
            'ó', 'Ó', '\xf3', '\xd3',
            'ć', 'Ć', '\u0107', '\u0106',
            'ś', 'Ś', '\u015b', '\u015a',
            'ż', 'Ż', '\u017c', '\u017b',
            'ź', 'Ź', '\u017a', '\u0179',
            'ń', 'Ń', '\u0144', '\u0143',
        ]

        for w, n in zip(weird, normal):
            text = text.replace(w, n)

        return text

    #
    # HASHTAGS
    #
    def transform_hashtags(self, text):
        def replace_hashtag(match):

            return "{prefix}{escape_seq}{text}".format(
                prefix=match.group('prefix'),
                escape_seq=HASHTAG_ESCAPE_SEQUENCE,
                text=match.group('text'))

        return HASHTAG_PATTERN.sub(replace_hashtag, text)

    #
    # LATEX
    #
    def transform_latex(self, conf, text):
        return transform_latex(text, conf)

    #
    # TOKENIZE
    #
    def tokenize(self, conf, text):
        with connection.cursor() as c:
            # deal with accents
            c.execute(
                "SELECT token FROM ts_debug(%s, %s) WHERE alias != 'blank'",
                [conf, text])

            tokens = [
                (t[0], i + 1)
                for i, t in enumerate(c.fetchall())]

        return [
            (token, position)
            for token, position in tokens
            if not stopwords_filter.is_stopword(conf, token)
        ]

    #
    # UNACCENTS
    #
    def augument_with_unaccents(self, conf, weight, stems, tokens):

        # -- unaccent original tokens
        for token, position in tokens:

            token = token.lower()
            unaccented_token = trans(token)
            if unaccented_token != token:
                stems.setdefault(unaccented_token, set())
                stems[unaccented_token] |= set([
                    self.get_position(position, weight)
                ])

        # -- unaccent stems
        for stem, positions in deepcopy(stems).items():

            stem = stem.lower()
            unaccented_stem = trans(stem)
            if unaccented_stem != stem:
                stems.setdefault(unaccented_stem, set())
                stems[unaccented_stem] |= positions

        return stems

    #
    # STEMS
    #
    def augument_with_stems(self, conf, text, weight, stems, tokens):

        if conf == 'polish':
            return self._augument_with_stems_polish(
                conf, weight, stems, tokens)

        else:
            return self._augument_with_stems_generic(conf, text, weight, stems)

    def _augument_with_stems_polish(self, conf, weight, stems, tokens):

        if conf == 'polish':
            for token, position in tokens:
                token_stems = self._get_polish_stems(token)
                for stem in token_stems:
                    stems.setdefault(stem, set())
                    stems[stem].add(self.get_position(position, weight))

        return stems

    def _get_polish_stems(self, token):
        if not self.hobj_pl:
            return []

        token_stems = self.hobj_pl.stem(token)
        stems = []

        if token_stems:
            for stem in token_stems:
                stem = stem.decode(self.hobj_pl.get_dic_encoding())
                stems.append(stem.lower())

        else:
            stems.append(token.lower())

        return stems

    def _augument_with_stems_generic(self, conf, text, weight, stems):
        with connection.cursor() as c:
            if not weight:
                c.execute("SELECT to_tsvector(%s, %s)", [conf, text])

            else:
                c.execute(
                    "SELECT setweight(to_tsvector(%s, %s), %s)",
                    [conf, text, weight])

            tsstems = self.from_tsvestor(c.fetchone()[0])

        for stem, positions in tsstems.items():
            stems.setdefault(stem, set())
            stems[stem] = stems[stem] | positions

        return stems

    #
    # GENERAL
    #
    def get_position(self, i, weight):
        if weight:
            return f'{i}{weight}'

        else:
            return f'{i}'

    def from_tsvestor(self, tsvector):
        stems = {}
        for stem_pos in tsvector.split():
            stem, positions = stem_pos.split(':')

            stem = stem.strip()
            stem = re.sub(r"^'", '', stem)
            stem = re.sub(r"'$", '', stem)

            stems[stem] = set(positions.split(','))

        return stems

    def to_tsvector(self, stems):
        def join_positions(x):
            return ','.join([str(c) for c in x])

        return ' '.join([
            f"'{stem}':{join_positions(positions)}"
            for stem, positions in stems.items()
        ])


#
# HELPERS
#
def _concatenate_tsvectors(v1, v2):

    with connection.cursor() as c:
        c.execute("SELECT %s::tsvector || %s::tsvector", [v1, v2])

        return c.fetchone()[0]


def concatenate_tsvectors(*vectors):

    v1 = vectors[0]
    for v in vectors[1:]:
        v1 = _concatenate_tsvectors(v1, v)

    return v1


class VectorCombinable(SearchVectorCombinable):

    def _combine(self, other, connector, reversed, node=None):
        return CombinedVector(self, connector, other, self.config)


class CombinedVector(VectorCombinable, CombinedExpression):
    def __init__(self, lhs, connector, rhs, config, output_field=None):
        self.config = config
        super(
            CombinedVector,
            self).__init__(lhs, connector, rhs, output_field)


class StoredVector(SearchVector, VectorCombinable):
    """Represent stored text-search vector.

    Used for wrapping either stored ``tsvector`` column or the aggregate of
    such columns.

    """

    def as_sql(self, compiler, connection, function=None, template=None):

        return super(StoredVector, self).as_sql(
            compiler,
            connection,
            function=None,
            template='%(expressions)s')


class OnTheFlyVector(SearchVector, VectorCombinable):
    """Represent on-the-fly calculated text-search vector.

    Used for parsing supplied column in on the fly fashion to ``tsvector``
    and optionally to combine itself with other tsvectors (either on the fly
    ones or stored ones).

    """




from unittest.mock import Mock

from django.test import TestCase
import pytest

from lily.base.name import (
    to_past,
    to_plural,
    BaseVerb,
    ConstantName,
    Execute,
    # -- CRUD
    Create,
    Read,
    Update,
    Delete,
    # -- BULK CRUD
    BulkCreate,
    BulkRead,
    BulkUpdate,
    BulkDelete,
    # -- CONDITIONAL CRUD
    CreateOrUpdate,
    CreateOrRead,
)
from lily.base.events import EventFactory


class ConstantNameTestCase(TestCase):

    def test_render_command_name(self):
        c = ConstantName('hi  there')

        assert c.render_command_name() == 'HI_THERE'

    def test_render_event_name(self):
        c = ConstantName('hi  there')

        assert c.render_event_name(
            Mock(), Mock(event='hello yo')) == 'HELLO_YO'


class VerbsTestCase(TestCase):

    def test_accepts_string_noun_or_model(self):

        card = Mock(_meta=Mock(model_name='Card'))
        user = Mock(_meta=Mock(model_name='User'))

        class GreetVerb(BaseVerb):
            verb = 'greet'

        # -- noun is the string
        v = GreetVerb('laundry')

        assert v.render_command_name() == 'GREET_LAUNDRY'

        # -- noun is a model
        v = GreetVerb(user)

        assert v.render_command_name() == 'GREET_USER'

        # -- noun is a model - triangulation
        v = GreetVerb(card)

        assert v.render_command_name() == 'GREET_CARD'

    def test_render_event_name__wrong_finalizer(self):

        # -- Success Executed of Created
        try:
            assert Create('Car').render_event_name(
                Mock(), EventFactory.Executed())

        except EventFactory.BrokenRequest as e:
            assert e.event == (
                'INVALID_FINALIZER_USED_FOR_SPECIFIC_COMMAND_DETECTED')

        # -- Updated instead of Deleted
        try:
            assert Delete('Car').render_event_name(
                Mock(), EventFactory.Updated())

        except EventFactory.BrokenRequest as e:
            assert e.event == (
                'INVALID_FINALIZER_USED_FOR_SPECIFIC_COMMAND_DETECTED')


class ExecuteTestCase(TestCase):

    def test_accepts_custom_verb(self):

        assert Execute('Get', 'Car').render_command_name() == 'GET_CAR'
        assert Execute('Get', 'Car').render_event_name() == 'CAR_GOT'

    def test_render_event_name__wrong_finalizer(self):

        try:
            assert Execute('Get', 'Car').render_event_name(
                Mock(), EventFactory.Created())

        except EventFactory.BrokenRequest as e:
            assert e.event == (
                'INVALID_FINALIZER_USED_FOR_SPECIFIC_COMMAND_DETECTED')


class CreateOrUpdateTestCase(TestCase):

    def test_render_event_name__accepts_many_finalizers(self):

        assert CreateOrUpdate('Car').render_event_name(
            Mock(), EventFactory.Created()) == 'CAR_CREATED'

        assert CreateOrUpdate('Box').render_event_name(
            Mock(), EventFactory.Updated()) == 'BOX_UPDATED'


class CreateOrReadTestCase(TestCase):

    def test_render_event_name__accepts_many_finalizers(self):

        assert CreateOrRead('Car').render_event_name(
            Mock(), EventFactory.Created()) == 'CAR_CREATED'

        assert CreateOrRead('Box').render_event_name(
            Mock(), EventFactory.Read()) == 'BOX_READ'


class BulkReadTestCase(TestCase):

    def test_applies_pluralization(self):

        user = Mock(_meta=Mock(model_name='User'))

        # -- render_command_name
        assert BulkRead('Car').render_command_name() == 'BULK_READ_CARS'

        assert BulkRead('boxes').render_command_name() == 'BULK_READ_BOXES'

        assert BulkRead(user).render_command_name() == 'BULK_READ_USERS'

        # -- render_event_name
        assert BulkRead('Car').render_event_name(
            Mock(), EventFactory.BulkRead()) == 'CARS_BULK_READ'

        assert BulkRead('boxes').render_event_name(
            Mock(), EventFactory.BulkRead()) == 'BOXES_BULK_READ'

        assert BulkRead(user).render_event_name(
            Mock(), EventFactory.BulkRead()) == 'USERS_BULK_READ'


@pytest.mark.parametrize(
    'verb, expected_command, expected_event', [
        # -- CRUD
        (Create('home'), 'CREATE_HOME', 'HOME_CREATED'),
        (Update('candy'), 'UPDATE_CANDY', 'CANDY_UPDATED'),
        (Read('bicycle'), 'READ_BICYCLE', 'BICYCLE_READ'),
        (Delete('home'), 'DELETE_HOME', 'HOME_DELETED'),
        # -- BULK CRUD
        (BulkCreate('box'), 'BULK_CREATE_BOXES', 'BOXES_BULK_CREATED'),
        (BulkRead('box'), 'BULK_READ_BOXES', 'BOXES_BULK_READ'),
        (BulkUpdate('lock'), 'BULK_UPDATE_LOCKS', 'LOCKS_BULK_UPDATED'),
        (BulkDelete('cat'), 'BULK_DELETE_CATS', 'CATS_BULK_DELETED'),
        (Execute('buy', 'home'), 'BUY_HOME', 'HOME_BOUGHT'),
        (CreateOrUpdate('home'), 'CREATE_OR_UPDATE_HOME', 'HOME_CREATED'),
        (CreateOrRead('home'), 'READ_OR_CREATE_HOME', 'HOME_CREATED'),
    ])
def test_verbs(verb, expected_command, expected_event):

    assert verb.render_command_name() == expected_command

    e = verb.finalizers[0]()
    assert verb.render_event_name(Mock(), e) == expected_event


@pytest.mark.parametrize(
    'verb, expected', [
        # -- irregular
        ('come', 'came'),
        ('pay', 'paid'),

        # -- single form
        ('read', 'read'),

        # -- regular
        ('list', 'listed'),

        # -- finishes with `e`
        ('create', 'created'),
        ('update', 'updated'),
        ('delete', 'deleted'),
        ('execute', 'executed'),

        # -- finishes with `y`
        ('execute', 'executed'),

        # -- with double consonant
        ('stop', 'stopped'),

        # -- exceptions
        ('refer', 'referred'),

        # -- NO double consonant
        ('visit', 'visited'),
    ])
def test_to_past(verb, expected):

    assert to_past(verb) == expected


@pytest.mark.parametrize(
    'noun, expected', [
        # -- regular
        ('apple', 'apples'),
        ('card', 'cards'),

        # -- irregular
        ('wolf', 'wolves'),

        # -- plural stay plural
        ('paths', 'paths'),
        ('wolves', 'wolves'),
    ])
def test_to_plural(noun, expected):

    assert to_plural(noun) == expected


@pytest.mark.parametrize(
    'phrase, expected_command, expected_event', [
        (
            Create('card').after.Delete('user'),
            'CREATE_CARD_AFTER_DELETE_USER',
            'CARD_CREATED_AFTER_USER_DELETED',
        ),

        (
            BulkRead('payment').after.CreateOrUpdate('card'),
            'BULK_READ_PAYMENTS_AFTER_CREATE_OR_UPDATE_CARD',
            'PAYMENTS_BULK_READ_AFTER_CARD_CREATE_OR_UPDATED',
        ),
    ])
def test_cause_and_effect(phrase, expected_command, expected_event):

    assert phrase.render_command_name() == expected_command

    e = phrase.effect.finalizers[0]()
    assert phrase.render_event_name(Mock(), e) == expected_event




import os
import tempfile
import textwrap
from unittest.mock import call

from django.test import TestCase, override_settings
import pytest
from lily_assistant.repo.repo import Repo

from lily.docs.renderers.angular.repo import (
    AngularHTTPRepo,
    AngularRepo,
    TemplateRepo,
    PathRule,
)


class AngularHTTPRepoTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    #
    # LINK
    #
    def test_link(self):
        npm = self.mocker.patch.object(AngularHTTPRepo, 'npm')
        r = AngularHTTPRepo()

        r.link()

        assert npm.call_args_list == [call('link')]


class AngularRepoTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.current_cwd = os.getcwd()

    def tearDown(self):
        os.chdir(self.current_cwd)

    #
    # GIT
    #
    def test_commit(self):
        git = self.mocker.patch.object(AngularRepo, 'git')
        r = AngularRepo('origin')

        r.commit('0.9.11')

        assert git.call_args_list == [call('commit -m "ADDED version 0.9.11"')]

    #
    # NPM
    #
    def test_build(self):
        npm = self.mocker.patch.object(AngularRepo, 'npm')
        r = AngularRepo('origin')

        r.build()

        assert npm.call_args_list == [call('run build')]

    def test_install(self):
        npm = self.mocker.patch.object(AngularRepo, 'npm')
        r = AngularRepo('origin')

        r.install()

        assert npm.call_args_list == [call('install')]

    def test_npm(self):
        execute = self.mocker.patch.object(AngularRepo, 'execute')
        r = AngularRepo('origin')

        r.npm('whatever')

        assert execute.call_args_list == [call('npm whatever')]

    #
    # UPGRADE VERSION
    #
    def test_upgrade_version(self):

        class MockConfig:

            @property
            def version(self):
                return '0.9.1'

        temp_dir = self.tmpdir.mkdir('repo')
        package = temp_dir.join('package.json')
        package.write('"version": "{}"'.format('0.1.18'))

        self.mocker.patch.object(
            tempfile, 'mkdtemp').return_value = str(temp_dir)

        r = AngularRepo('origin')
        r.cd_to_repo()

        assert r.upgrade_version(MockConfig()) == '0.9.1-API-0.1.18-CLIENT'
        with open('package.json') as f:
            assert f.read() == '"version": "0.9.1-API-0.1.18-CLIENT"'


class PathRuleTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, tmpdir):
        self.tmpdir = tmpdir

    #
    # MATCHES
    #
    def test_matches__file(self):

        rule = PathRule(r'.*\.json$')

        assert rule.matches('/hi/there.json') is True
        assert rule.matches('/hi/there.json123') is False
        assert rule.matches('woeldOfjson') is False
        assert rule.matches('this.JSON') is True

    def test_matches__directory(self):

        base_dir = self.tmpdir
        base_dir.mkdir('modules')

        rule = PathRule(r'.*modules', True)

        assert rule.matches(str(base_dir.join('modules'))) is True
        assert rule.matches(str(base_dir.join('modules/here'))) is True
        assert rule.matches(str(base_dir.join('modules123'))) is False
        assert rule.matches(str(base_dir.join('modles'))) is False

    def test_matches__matched_but_not_directory(self):

        file = self.tmpdir.mkdir('hello').join('modules').write('hi')

        rule = PathRule(r'.*modules')

        assert rule.matches(str(file)) is False


class TemplateRepoTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    #
    # CLONE
    #
    @override_settings(LILY_ANGULAR_CLIENT_ORIGIN='some_origin')
    def test_clone__makes_the_right_calls(self):

        clone = self.mocker.patch.object(Repo, 'clone')
        temp_dir = self.tmpdir.mkdir('123')
        self.mocker.patch.object(
            tempfile, 'mkdtemp').return_value = str(temp_dir)
        repo = TemplateRepo()

        repo.clone()

        assert clone.call_args_list == [call(str(temp_dir))]

    #
    # COPY_TO
    #
    def test_copy_to__ignores_and_keeps(self):

        source_dir = self.tmpdir.mkdir('source')
        d_0 = source_dir.mkdir('d_0')

        f_0_0 = d_0.join('f_0_0.md')
        f_0_0.write('f_0_0')

        f_0_1 = d_0.join('f_0_1.js')  # noqa - will be ignored
        f_0_1.write('f_0_1')

        d_0_0 = d_0.mkdir('d_0_0')

        f_0_0_0 = d_0_0.join('f_0_0_0.json')
        f_0_0_0.write('f_0_0_0')

        f_0_0_1 = d_0_0.join('f_0_0_1.ts')
        f_0_0_1.write('f_0_0_1')

        d_1 = source_dir.mkdir('d_1')

        f_1_0 = d_1.join('f_1_0.ico')
        f_1_0.write('f_1_0')

        self.mocker.patch.object(
            tempfile, 'mkdtemp').return_value = str(source_dir)

        destination_dir = self.tmpdir.mkdir('destination')
        repo = TemplateRepo()

        repo.copy_to(str(destination_dir), 'extra')

        assert set(os.listdir(destination_dir)) == set(['d_0', 'd_1'])
        assert (
            set(os.listdir(destination_dir.join('d_0'))) ==
            set(['f_0_0.md', 'd_0_0']))
        assert (
            set(os.listdir(destination_dir.join('d_0/d_0_0'))) ==
            set(['f_0_0_0.json', 'f_0_0_1.ts']))
        assert (
            set(os.listdir(destination_dir.join('d_1'))) ==
            set(['f_1_0.ico']))

    def test_copy_to__cleans_up_before_copy(self):

        source_dir = self.tmpdir.mkdir('source')
        d = source_dir.mkdir('d')
        f = d.join('f.md')
        f.write('f')

        destination_dir = self.tmpdir.mkdir('destination')

        # -- should be kept
        git_dir = destination_dir.mkdir('.git')
        git_dir.join('hook').write('hook it')

        # -- should be removed
        dd = destination_dir.mkdir('dd')
        dd.join('hi.txt').write('hello')
        destination_dir.join('root.js').write('root it')

        self.mocker.patch.object(
            tempfile, 'mkdtemp').return_value = str(source_dir)

        repo = TemplateRepo()

        repo.copy_to(str(destination_dir), 'extra')

        assert set(os.listdir(destination_dir)) == set(['d', '.git'])
        assert os.listdir(destination_dir.join('d')) == ['f.md']
        assert os.listdir(destination_dir.join('.git')) == ['hook']

    #
    # IGNORE
    #
    def test_ignore__ignores(self):

        repo = TemplateRepo()

        self.tmpdir.mkdir('node_modules')
        self.tmpdir.mkdir('node_modules/hello')
        self.tmpdir.mkdir('.git')
        self.tmpdir.mkdir('hi')

        to_ignore = repo.ignore(
            str(self.tmpdir),
            [
                # -- ignore: node_modules and all its children
                'node_modules',
                'node_modules/hello',
                'node_modules/angular.ts',

                # -- ignore: .git and all its children
                '.git',
                '.git/hooks',
                '.git/whatever.json',

                # -- no ignore .ts
                'hello.ts',
                'hi/hello.ts',
            ])

        assert set(to_ignore) == set([
            '.git',
            '.git/hooks',
            '.git/whatever.json',

            'node_modules',
            'node_modules/hello',
            'node_modules/angular.ts',
        ])

    def test_ignore__keeps(self):

        repo = TemplateRepo()

        self.tmpdir.mkdir('x')
        self.tmpdir.mkdir('x/y')

        to_ignore = repo.ignore(
            str(self.tmpdir),
            [
                # -- exceptions
                'a/karma.conf.js',
                'a/browserslist',
                'a/.npmignore',
                'a/.gitkeep',

                # -- ts
                'a/hi.ts',
                'a/b/service.ts',

                # -- html
                'b/app.html',
                'b/c/what.html',

                # -- json
                'hi.json',
                'c/wat.json',

                # -- ico
                'hi.ico',
                'c/wat.ico',

                # -- css
                'hi.css',
                'c/wat.css',

                # -- all other directories
                'x/what.json',
                'x/y/whatever.ts',
            ])

        # -- all to keep
        assert to_ignore == []

    #
    # COPY
    #
    def test_copy(self):

        source = self.tmpdir.join('hello.ts')
        source.write('hello there')

        destination = self.tmpdir.join('copy_hello.ts')

        repo = TemplateRepo()

        repo.copy('super', str(source), str(destination))

        assert destination.read() == 'hello there'

    def test_copy__replaces_prefix(self):

        source = self.tmpdir.join('hello.ts')
        source.write(textwrap.dedent('''
            hello __CLIENT_PREFIX__

            __CLIENT_PREFIX__ is cool.
        '''))

        destination = self.tmpdir.join('copy_hello.ts')

        repo = TemplateRepo()

        repo.copy('super', str(source), str(destination))

        assert destination.read() == textwrap.dedent('''
            hello super

            super is cool.
        ''')

    def test_copy__deals_with_binary_files(self):

        source = self.tmpdir.join('hello.ico')
        source.write(b'\x00\x00\x01\x00\x02\x00\x10\x10\x00\x00', 'wb')

        destination = self.tmpdir.join('copy_hello.ico')

        repo = TemplateRepo()

        repo.copy('super', str(source), str(destination))

        assert destination.read() == '\x00\x00\x01\x00\x02\x00\x10\x10\x00\x00'




import os
from unittest.mock import call, Mock
import importlib
import json

from django.test import TestCase
from click.testing import CliRunner
import pytest

from migrator.management.commands.render_current_migration_plan import (
    Renderer,
    command,
)


def migration_class(dependencies=None):

    return Mock(Migration=Mock(dependencies=dependencies or []))


class RendererTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()
        self.renderer = Renderer()

        class MockConfig:

            @property
            def src_dir(self):
                return 'src'

        self.mocker.patch(
            'migrator.management.commands.render_current_migration_plan'
            '.Config',
            MockConfig)

        self.prev_dir = os.getcwd()
        srcdir = self.tmpdir.mkdir('src')

        os.chdir(self.tmpdir)

        # -- BUILD BASIC STRUCTURE
        self.auth_dir = srcdir.mkdir('auth').mkdir('migrations')
        self.task_dir = srcdir.mkdir('task').mkdir('migrations')
        self.todo_dir = srcdir.mkdir('todo').mkdir('migrations')

    def tearDown(self):
        os.chdir(self.prev_dir)

    #
    # RENDER
    #
    def test_render__makes_the_right_calls(self):

        get_dependencies = self.mocker.patch.object(
            Renderer, 'get_dependencies')
        get_dependencies.return_value = {
            ('app0', '001_what'): set(['what']),
            ('app1', '006_why'): set(),
        }
        get_most_current_migrations = self.mocker.patch.object(
            Renderer, 'get_most_current_migrations')
        get_migrations_plan = self.mocker.patch.object(
            Renderer, 'get_migrations_plan')

        the_plan = self.renderer.render()

        assert the_plan == {
            'plan': get_migrations_plan.return_value,
            'dependencies': {
                'app0.001_what': ['what'],
                'app1.006_why': [],
            }
        }
        assert get_dependencies.call_args_list == [
            call(get_most_current_migrations.return_value),
        ]
        assert get_migrations_plan.call_args_list == [
            call(get_dependencies.return_value),
        ]

    #
    # GET_MOST_CURRENT_MIGRATIONS
    #
    def test_get_most_current_migrations__no_migrations(self):

        self.renderer.get_most_current_migrations() == []

    def test_get_most_current_migrations__only_current_migrations(self):

        self.auth_dir.join('0001_add_column.py').write(migration_class())
        self.task_dir.join('0001_add_model.py').write(migration_class())

        assert self.renderer.get_most_current_migrations() == [
            (
                'task',
                '0001_add_model',
                'src.task.migrations.0001_add_model',
            ),
            (
                'auth',
                '0001_add_column',
                'src.auth.migrations.0001_add_column',
            ),
        ]

    def test_get_most_current_migrations__not_only_current_migrations(self):

        self.auth_dir.join('0001_add_column.py').write(migration_class())
        self.auth_dir.join('0002_remove_column.py').write(migration_class())
        self.auth_dir.join('0003_rename_column.py').write(migration_class())

        self.task_dir.join('0001_add_model.py').write(migration_class())
        self.task_dir.join('0002_add_column.py').write(migration_class())

        assert self.renderer.get_most_current_migrations() == [
            (
                'task',
                '0002_add_column',
                'src.task.migrations.0002_add_column',
            ),
            (
                'auth',
                '0003_rename_column',
                'src.auth.migrations.0003_rename_column',
            ),
        ]

    #
    # GET_DEPENDENCIES
    #
    def test_get_dependencies__no_dependencies(self):

        self.mocker.patch.object(importlib, 'import_module').side_effect = [
            migration_class(),
            migration_class(),
        ]

        assert self.renderer.get_dependencies([
            ('auth', '0001_add_column', 'auth.migrations.0001_add_column'),
            ('task', '0002_add_column', 'task.migrations.0002_add_column'),
        ]) == {
            ('auth', '0001_add_column'): set(),
            ('task', '0002_add_column'): set(),
        }

    def test_get_dependencies__some_dependencies__between_each_other(self):

        self.mocker.patch.object(importlib, 'import_module').side_effect = [
            migration_class([
                ('task', '0002_add_record'),
                ('todo', '0011_move_record'),
            ]),
            migration_class(),
            migration_class([
                ('task', '0002_add_record'),
            ]),
        ]

        assert self.renderer.get_dependencies([
            ('auth', '0001_add_record', 'auth.migrations.0001_add_record'),
            ('task', '0002_add_record', 'task.migrations.0002_add_record'),
            ('todo', '0011_move_record', 'todo.migrations.0011_move_record'),
        ]) == {
            ('auth', '0001_add_record'): set([
                ('task', '0002_add_record'),
                ('todo', '0011_move_record'),
            ]),
            ('task', '0002_add_record'): set(),
            ('todo', '0011_move_record'): set([
                ('task', '0002_add_record'),
            ]),
        }

    def test_get_dependencies__some_dependencies__some_external(self):

        self.mocker.patch.object(importlib, 'import_module').side_effect = [
            migration_class([
                ('task', '0002_add_row'),
                ('product', '00038_add_model'),
            ]),
            migration_class(),
            migration_class([
                ('task', '0002_add_row'),
            ]),
        ]

        assert self.renderer.get_dependencies([
            ('auth', '0001_add_row', 'auth.migrations.0001_add_row'),
            ('task', '0002_add_row', 'task.migrations.0002_add_row'),
            ('todo', '0011_move_row', 'todo.migrations.0011_move_row'),
        ]) == {
            ('auth', '0001_add_row'): set([
                ('task', '0002_add_row'),
                ('product', '00038_add_model'),
            ]),
            ('task', '0002_add_row'): set(),
            ('todo', '0011_move_row'): set([
                ('task', '0002_add_row'),
            ]),
        }

    #
    # GET_MIGRATIONS_PLAN
    #
    def test_get_migrations_plan__no_migrations(self):

        assert self.renderer.get_migrations_plan({}) == []

    def test_get_migrations_plan__single_migration_with_prev_dependency(self):

        assert self.renderer.get_migrations_plan({
            ('search', '0002_a'): set([
                ('search', '0001_a'),
            ]),
        }) == [
            ('search', '0002_a'),
        ]

    def test_get_migrations_plan__no_dependencies(self):

        assert self.renderer.get_migrations_plan({
            ('auth', '0001_add_column'): set(),
            ('task', '0002_add_column'): set(),
            ('todo', '0011_move_column'): set(),
        }) == [
            ('auth', '0001_add_column'),
            ('task', '0002_add_column'),
            ('todo', '0011_move_column'),
        ]

    def test_get_migrations_plan__some_dependencies(self):

        assert self.renderer.get_migrations_plan({
            ('auth', '0001_c'): set([
                ('task', '0002_c'),
                ('todo', '0011_c'),
                ('joke', '0001_c'),
            ]),
            ('task', '0002_c'): set(),
            ('todo', '0011_c'): set([
                ('task', '0002_c'),
            ]),
        }) == [
            ('task', '0002_c'),
            ('todo', '0011_c'),
            ('auth', '0001_c'),
        ]


class RenderCurrentMigrationPlanCommandTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()

        class MockConfig:

            @property
            def version(self):
                return '0.9.1'

            @property
            def next_version(self):
                return None

            @classmethod
            def get_lily_path(cls):
                return str(self.tmpdir)

        self.mocker.patch(
            'migrator.management.commands.render_current_migration_plan'
            '.Config',
            MockConfig)

    def test_migrations_directory_created_with_migration_file(self):

        self.mocker.patch.object(Renderer, 'render').return_value = {
            'some': 'thing',
        }
        result = self.runner.invoke(command)

        assert result.exit_code == 0
        assert result.output.strip() == (
            f'Migrations plan rendered for to file {str(self.tmpdir)}'
            f'/migrations/0.9.1.json')
        assert self.tmpdir.listdir() == [self.tmpdir.join('migrations')]
        assert self.tmpdir.join('migrations').listdir() == [
            self.tmpdir.join('migrations').join('0.9.1.json')
        ]
        assert json.loads(
            self.tmpdir.join('migrations').join('0.9.1.json').read('r')
        ) == {
            'some': 'thing',
        }

    def test_migrations_directory_already_exists(self):

        self.tmpdir.mkdir('migrations').join('0.9.0.json').write('{}')
        self.mocker.patch.object(Renderer, 'render').return_value = {
            'some': 'thing',
        }
        result = self.runner.invoke(command)

        assert result.exit_code == 0
        assert result.output.strip() == (
            f'Migrations plan rendered for to file {str(self.tmpdir)}'
            f'/migrations/0.9.1.json')
        assert self.tmpdir.listdir() == [self.tmpdir.join('migrations')]
        assert self.tmpdir.join('migrations').listdir() == [
            self.tmpdir.join('migrations').join('0.9.0.json'),
            self.tmpdir.join('migrations').join('0.9.1.json'),
        ]
        assert json.loads(
            self.tmpdir.join('migrations').join('0.9.1.json').read('r')
        ) == {
            'some': 'thing',
        }




import os

import yaml
from langid.langid import LanguageIdentifier, model

from lily.base.events import EventFactory


def load_from_data(filename):

    filepath = os.path.join(
        os.path.dirname(__file__),
        'data',
        filename)

    return yaml.load(open(filepath, 'r'), Loader=yaml.FullLoader)


LANGUAGE_TO_CONFIGURATION = load_from_data(
    'language_to_configuration_map.yaml')


ALL_LANGUAGES = load_from_data(
    'language_codes.yaml')


class LanguageDetector(EventFactory):

    DETECT_THRESHOLD_LEN = 3
    """
    Consider maximum 3 detected languages.

    """

    DETECT_MIN_LANG_LENGTH = 5
    """
    Minimum length of text which will trigger the language detection mechanism.
    Every text with length below this threshold automatically is treated as
    following ``simple`` configuration.

    """

    DETECT_THRESHOLD_PROB = 0.01
    """
    Consider only those languages that have rank probability higher that 0.3

    """

    # -- languages supported by langid
    language_abbrs = set([
        'af', 'am', 'an', 'ar', 'as', 'az', 'be', 'bg', 'bn', 'br', 'bs',
        'ca', 'cs', 'cy', 'da', 'de', 'dz', 'el', 'en', 'eo', 'es', 'et',
        'eu', 'fa', 'fi', 'fo', 'fr', 'ga', 'gl', 'gu', 'he', 'hi', 'hr',
        'ht', 'hu', 'hy', 'id', 'is', 'it', 'ja', 'jv', 'ka', 'kk', 'km',
        'kn', 'ko', 'ku', 'ky', 'la', 'lb', 'lo', 'lt', 'lv', 'mg', 'mk',
        'ml', 'mn', 'mr', 'ms', 'mt', 'nb', 'ne', 'nl', 'nn', 'no', 'oc',
        'or', 'pa', 'pl', 'ps', 'pt', 'qu', 'ro', 'ru', 'rw', 'se', 'si',
        'sk', 'sl', 'sq', 'sr', 'sv', 'sw', 'ta', 'te', 'th', 'tl', 'tr',
        'ug', 'uk', 'ur', 'vi', 'vo', 'wa', 'xh', 'zh', 'zu'
    ])

    identifier = None

    def __init__(self):

        self.languages = [
            lang
            for lang in ALL_LANGUAGES
            if lang['abbr'] in self.language_abbrs]

        self.languages_index = {
            lang['abbr']:
            lang for lang in self.languages
        }

    def detect(self, text):

        if not LanguageDetector.identifier:
            LanguageDetector.identifier = LanguageIdentifier.from_modelstring(
                model, norm_probs=True)

        language_abbrs = [
            lang
            for lang, prob in LanguageDetector.identifier.rank(text)
            if prob > self.DETECT_THRESHOLD_PROB
        ][:self.DETECT_THRESHOLD_LEN]

        if not language_abbrs:
            raise self.BrokenRequest(
                'UNSUPPORTED_LANGUAGE_DETECTED',
                data={'text': text},
                is_critical=True)

        return [
            self.languages_index[abbr]
            for abbr in language_abbrs
            if abbr in self.languages_index]

    def detect_db_conf(self, text):

        text = text.strip()
        if len(text) < self.DETECT_MIN_LANG_LENGTH:
            return 'simple'

        try:
            detected = self.detect(text)[0]['abbr']

            return LANGUAGE_TO_CONFIGURATION[detected]['configuration']

        except (KeyError, IndexError, self.BrokenRequest):
            return 'simple'


detector = LanguageDetector()




import re
import os
from copy import copy, deepcopy

from rest_framework import serializers as drf_serializers
from rest_framework.serializers import (  # noqa
    BooleanField,
    CharField,
    ChoiceField,
    DateField,
    DateTimeField,
    DecimalField,
    DictField,
    EmailField,
    FloatField,
    IntegerField,
    JSONField,
    ListField,
    ListSerializer,
    SerializerMethodField,
    URLField,
    UUIDField,
    NullBooleanField,
    ValidationError,
    ReadOnlyField,
)

from .events import EventFactory


BASE_DIR = os.path.dirname(__file__)


COMMANDS_CONF = {}


class MissingTypeError(Exception):
    """Could not find `_type` attribute.

    When Serializer is missing the semantic `_type` denoting a type which
    Entity a given Serializer represents.

    """


class MissingRequiredArgumentsException(Exception):
    pass


class EnumChoiceField(ChoiceField):

    def __init__(self, *args, enum_name=None, enum=None, **kwargs):

        if enum_name:
            self.enum_name = enum_name

        if enum:
            self.enum_name = enum.__name__
            kwargs['choices'] = [e.value for e in enum]

        if not (enum_name or enum):
            raise MissingRequiredArgumentsException(
                'either `enum_name` or `enum` must be provided')

        super(EnumChoiceField, self).__init__(*args, **kwargs)


class Serializer(drf_serializers.Serializer, EventFactory):

    def __init__(self, *args, fields_subset=None, **kwargs):
        self._fields_subset = fields_subset
        super(Serializer, self).__init__(*args, **kwargs)

    def render_access(self, instance):

        try:
            out_access = {}
            for cmd, is_active in self.get_access(instance):
                if isinstance(cmd, str):
                    name = cmd

                else:
                    name = cmd.command_conf['name']

                out_access[name] = is_active

            return out_access

        except AttributeError:
            pass

        return {}

    def to_internal_value(self, data):

        try:
            transformed = deepcopy(data)
            for field_name, value in data.items():
                if field_name.startswith('@'):
                    new_field_name = re.sub(r'^@', 'at__', field_name)
                    transformed[new_field_name] = value
                    del transformed[field_name]

            return super(Serializer, self).to_internal_value(transformed)

        except AttributeError:
            # -- HACK: which allows me to return instances of models inside
            # -- normal dictionary
            return data

    def to_representation(self, instance):

        if self._fields_subset:
            body = {
                f: getattr(instance, f)
                for f in self._fields_subset
            }

        # -- when dealing with DICT
        elif isinstance(instance, dict):
            transformed = copy(instance)
            for field_name, value in instance.items():
                if field_name.startswith('@'):
                    new_field_name = re.sub(r'^@', 'at__', field_name)
                    transformed[new_field_name] = value
                    del transformed[field_name]

            body = super(Serializer, self).to_representation(transformed)

        # -- when dealing normal instance
        else:
            body = super(Serializer, self).to_representation(instance)

        # -- transform `at__` to `@`
        original = copy(body)
        for field_name, value in original.items():
            if field_name.startswith('at__'):
                new_field_name = re.sub(r'^at__', '@', field_name)
                body[new_field_name] = value
                del body[field_name]

        # --
        # -- attach type meta info
        # --
        # -- for `AbstractSerializer` do not attach any extra data they are
        # -- here only for storing SubEntities without having their own
        # -- interpretation
        if isinstance(self, AbstractSerializer):
            return body

        try:
            body['@type'] = self._type

        except AttributeError:
            raise MissingTypeError(
                'Each Serializer must have `type` specified which informs '
                'the client about the semantic type a result of the '
                'Serializer represents')

        access = self.render_access(instance)
        if access:
            body['@access'] = access

        return body


class AbstractSerializer(Serializer):
    """Store abstract entities.

    Designed for storing only abstract entities without having it's
    own meaning.

    """


class ModelSerializer(drf_serializers.ModelSerializer, Serializer):

    serializer_choice_field = EnumChoiceField

    def build_standard_field(self, field_name, model_field):

        from . import models  # noqa - avoid circular dependency

        field_class, field_kwargs = super(
            ModelSerializer, self).build_standard_field(
                field_name, model_field)

        if isinstance(model_field, models.EnumChoiceField):
            field_kwargs['enum_name'] = model_field.enum_name
            field_class = EnumChoiceField

        return field_class, field_kwargs


class EmptySerializer(Serializer):

    _type = 'empty'


class ObjectSerializer(Serializer):

    _type = 'object'

    def to_internal_value(self, data):

        return data

    def to_representation(self, data):

        data['@type'] = self._type

        return data


class CommandSerializer(Serializer):

    _type = 'command'

    name = CharField()

    method = ChoiceField(choices=('post', 'get', 'delete', 'put'))

    uri = URLField()

    body = DictField(required=False)

    query = DictField(required=False)

    result = DictField(required=False)


class JSONSchemaField(JSONField):
    def __init__(self, *args, **kwargs):
        from .models import JSONSchemaValidator

        class JSONSchemaValidatorSerializer(JSONSchemaValidator):

            validation_error_cls = ValidationError

        try:
            schema = kwargs.pop('schema')

        except KeyError:
            schema = {}

        super(JSONSchemaField, self).__init__(*args, **kwargs)
        self.validators.insert(
            0, JSONSchemaValidatorSerializer(schema=schema))




import asyncio
import concurrent.futures

from .task import AsyncTask


class ParallelExecutor:

    BACKOFF_UNIT = 3  # seconds

    def __init__(self, tasks: [AsyncTask]):

        self.tasks = tasks

    async def async_execute(self, loop):

        with concurrent.futures.ThreadPoolExecutor() as pool:
            futures = [
                loop.run_in_executor(pool, task.callback, *task.args)
                for task in self.tasks]

        future_results = asyncio.gather(*futures, return_exceptions=True)
        for i, result in enumerate(await future_results):
            self.tasks[i].result = result
            self.tasks[i].successful = not isinstance(result, Exception)

        return self.tasks

    def execute(self):

        loop = asyncio.new_event_loop()
        result = loop.run_until_complete(self.async_execute(loop))
        loop.close()

        return result




from unittest import TestCase

import pytest

from lily.base.utils import normalize_indentation, Timer


def s(*args):
    return '\n'.join(args)


@pytest.mark.parametrize(
    'text, min_indent, expected',
    [
        # -- empty to empty
        ('', 0, ''),

        # -- single liner
        (
            '''
            hello world
            ''',
            0,
            s('hello world'),
        ),

        # -- trim beginning and ending
        (
            '''

            hello world


            ''',
            0,
            s('hello world'),
        ),

        # -- multi line
        (
            '''
            hello world
            or not hello
            ''',
            0,
            s(
                'hello world',
                'or not hello',
            ),
        ),

        # -- multi line - varying indent
        (
            '''
            hello world
                or not hello
                  what
            ''',
            0,
            s(
                'hello world',
                '    or not hello',
                '      what'
            ),
        ),

        # -- multi line - non zero min_indent
        (
            '''
            hello world
                or not hello
                  what
            ''',
            2,
            s(
                '  hello world',
                '      or not hello',
                '        what'
            ),
        ),

        # -- multi line - empty lines between
        (
            '''
            hello world


            what
            ''',
            0,
            s(
                'hello world',
                '',
                '',
                'what'
            ),
        ),

    ])
def test_normalize_indentation(text, min_indent, expected):
    assert normalize_indentation(text, min_indent) == expected


class TimerTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker):
        self.mocker = mocker

    #
    # ELASED
    #
    def test_elapsed(self):

        M = 60  # noqa
        H = 3600  # noqa

        self.mocker.patch('lily.base.utils.time').side_effect = [
            # -- 1st
            134,
            134 + 11.012,
            # -- 2nd
            146,
            146 + (3 * M + 17.189000),
            # -- 3rd
            189,
            189 + (2 * H + 7 * M + 17),
        ]

        with Timer() as t:
            assert t.elapsed == '00:00:11.012'

        with Timer() as t:
            assert t.elapsed == '00:03:17.189'

        with Timer() as t:
            assert t.elapsed == '02:07:17.000'




import json
import os

from lily_assistant.config import Config
from django.conf import settings

from lily.base.events import EventFactory
from .command import Command
from lily.base.utils import normalize_indentation
from .repo import AngularRepo, AngularHTTPRepo, TemplateRepo
from .domain import Domain


class local_cwd:  # noqa

    def __init__(self, cwd):
        self.cwd = cwd

    def __enter__(self):
        self.current_cwd = os.getcwd()
        os.chdir(self.cwd)

    def __exit__(self, *args, **kwargs):
        os.chdir(self.current_cwd)


class AngularClientRenderer(EventFactory):

    AUTOGENERATED_DOC = normalize_indentation('''
        /**
          * THIS FILE WAS AUTOGENERATED, ALL MANUAL CHANGES CAN BE
          * OVERWRITTEN
          */
    ''', 0)

    def __init__(self, client_origin, client_prefix):
        # -- source repo
        self.template_repo = TemplateRepo()

        self.http_repo = AngularHTTPRepo()

        # -- target repo
        self.repo = AngularRepo(client_origin)
        self.client_prefix = client_prefix

    #
    # MAIN
    #
    def render(
            self,
            only_build=False,
            include_domains=None,
            exclude_domains=None):

        root_cwd = os.getcwd()

        # -- save it now before all directory jumps
        config = Config()

        # -- HTTP client
        self.http_repo.cd_to_repo()
        self.http_repo.clone()
        self.http_repo.install()
        self.http_repo.build()
        self.http_repo.link()

        # -- pull newest changes to the template
        self.template_repo.clone()

        # -- copy its content to the temp directory into which client will get
        # -- render
        self.repo.clone()

        self.template_repo.copy_to(self.repo.base_path, self.client_prefix)
        self.repo.cd_to_repo()
        self.repo.install()
        self.repo.link()

        # -- render domains
        with local_cwd(root_cwd):
            commands_by_domain = self.group_commands_by_domain(
                include_domains, exclude_domains)

        # -- render client
        self.render_client_module_ts(commands_by_domain)
        self.render_api_ts(commands_by_domain)
        self.render_api_index_ts(commands_by_domain)

        # -- particular domains
        all_commands = []
        for domain, commands in commands_by_domain.items():
            sorted_commands = []
            for command_name in sorted(commands.keys()):
                sorted_commands.append(commands[command_name])

            self.render_domain(domain, sorted_commands)
            all_commands.extend(sorted_commands)

        # -- create SHARED folder
        rel_path = './projects/client/src/shared'
        self.repo.create_dir(rel_path)

        # -- access for all commands
        self.render_access_ts(rel_path, all_commands)

        # -- enums - must be run after domains rendering since it
        # -- renders enums
        self.render_enums_ts(rel_path, all_commands)

        self.repo.build()

        if not only_build:
            next_version = self.repo.upgrade_version(config)

            self.repo.add_all()
            self.repo.commit(next_version)
            self.repo.push()

            return next_version

    def group_commands_by_domain(
            self, include_domains=None, exclude_domains=None):

        include_domains = include_domains or []
        include_domains = [d.lower() for d in include_domains]
        exclude_domains = exclude_domains or []
        exclude_domains = [d.lower() for d in exclude_domains]

        commands_by_domain = {}
        for name, conf in self.get_commands().items():
            if name == '@enums':
                continue

            command = Command(name, conf)

            # -- skip excluded domains
            if command.domain_id.lower() in exclude_domains:
                continue

            # -- skip not included domains
            if (include_domains and
                    command.domain_id.lower() not in include_domains):
                continue

            domain = Domain(command.domain_id, command.domain_name)

            if not command.is_private:
                commands_by_domain.setdefault(domain, {})
                commands_by_domain[domain][name] = command

        return commands_by_domain

    def get_commands(self):

        commands_path = os.path.join(
            Config.get_lily_path(),
            'commands',
            f'{Config().version}.json')
        with open(commands_path, 'r') as f:
            return json.loads(f.read())

    #
    # Domain Specific Folder
    #
    def render_domain(self, domain, commands):

        self.repo.create_dir(domain.path.base_path)
        self.repo.clear_dir(domain.path.base_path)

        # -- RENDER INDEX TS
        self.render_index_ts(domain)

        # -- RENDER MODELS TS
        self.render_models_ts(domain, commands)

        # -- RENDER DOMAIN TS
        self.render_domain_ts(domain, commands)

        # -- RENDER EXAMPLES TS
        self.render_examples_ts(domain, commands)

    def render_index_ts(self, domain):
        with open(domain.path.join('index.ts'), 'w') as f:
            f.write(
                normalize_indentation('''
                    export * from './{domain.id}.domain';
                    export * from './{domain.id}.models';
                ''', 0).format(domain=domain))

    def render_models_ts(self, domain, commands):

        model_enums = []
        models = []
        for command in commands:
            # -- REQUEST_QUERY
            b, enums = command.request_query.render()
            if b:
                models.append(b)

            if enums:
                model_enums.extend(enums)

            # -- REQUEST_BODY
            b, enums = command.request_body.render()
            if b:
                models.append(b)

            if enums:
                model_enums.extend(enums)

            # -- RESPONSE
            b, enums = command.response.render()
            if b:
                models.append(b)

            if enums:
                model_enums.extend(enums)

        enums_block = ''
        if model_enums:
            sorted_enum_names = sorted(set([e.name for e in model_enums]))

            enums_block = normalize_indentation('''
                import {{ {enums} }} from '../../shared/enums';
            ''', 0).format(enums=', '.join(sorted_enum_names))

        blocks = [
            normalize_indentation('''
                {autogenerated_doc}

                {enums_block}

                /**
                 * {domain_name} Domain Models
                 */
            ''', 0).format(
                enums_block=enums_block,
                autogenerated_doc=self.AUTOGENERATED_DOC,
                domain_name=domain.name),
            *models,
        ]

        path = domain.path.join('{}.models.ts'.format(domain.id))
        with open(path, 'w') as f:
            f.write('\n\n'.join(blocks))

    def render_domain_ts(self, domain, commands):
        blocks = [
            normalize_indentation('''
                {autogenerated_doc}

                /**
                 * {domain.name} Domain
                 */
                import {{ Injectable }} from '@angular/core';
                import {{ filter }} from 'rxjs/operators';
                import {{ Observable }} from 'rxjs';
                import * as _ from 'underscore';

                import {{ HttpService }} from '../../http.service';

                import * as X from './{domain.id}.models';

                @Injectable()
                export class {domain.camel_id}Domain {{
                    constructor(private client: HttpService) {{}}
            ''', 0).format(  # noqa
                autogenerated_doc=self.AUTOGENERATED_DOC,
                domain=domain,
            )
        ]
        for command in commands:
            blocks.append(normalize_indentation(command.render(), 4))

        blocks.append('}')

        path = domain.path.join('{}.domain.ts'.format(domain.id))
        with open(path, 'w') as f:
            f.write('\n\n'.join(blocks))

    def render_examples_ts(self, domain, commands):
        blocks = [
            normalize_indentation('''
                {autogenerated_doc}

                /**
                 * {domain.name} Domain Examples
                 */
            ''', 0).format(  # noqa
                autogenerated_doc=self.AUTOGENERATED_DOC,
                domain=domain,
            )
        ]
        for command in commands:
            blocks.append(normalize_indentation(command.render_examples(), 0))

        path = domain.path.join('{}.examples.ts'.format(domain.id))
        with open(path, 'w') as f:
            f.write('\n\n'.join(blocks))

    def render_access_ts(self, rel_path, all_commands):
        header = normalize_indentation('''
            {autogenerated_doc}
        ''', 0).format(  # noqa
            autogenerated_doc=self.AUTOGENERATED_DOC)

        access_enum = settings.LILY_AUTHORIZER_ACCESS_ENUM_CLASS
        blocks = [
            command.render_access(access_enum) for command in all_commands]
        blocks = '\n'.join([f'  {b},' for b in blocks])

        abs_path = os.path.join(self.repo.base_path, rel_path)
        with open(os.path.join(abs_path, 'access.ts'), 'w') as f:
            f.write(
                f'{header}\n\n'
                f'import {{ { access_enum } }} from \'./enums\';\n\n'  # noqa
                f'export const Access = {{\n'
                f'{blocks}\n'
                '}')

    def render_enums_ts(self, rel_path, all_commands):

        enums = self.collect_unique_enums(all_commands)

        blocks = [enum.render() for enum in enums]

        abs_path = os.path.join(self.repo.base_path, rel_path)
        with open(os.path.join(abs_path, 'enums.ts'), 'w') as f:
            f.write('\n\n'.join(blocks))

    def collect_unique_enums(self, all_commands):

        all_enums = []

        for command in all_commands:
            command.request_query.render()
            command.request_body.render()
            command.response.render()

            all_enums.extend(command.request_query.enums)
            all_enums.extend(command.request_body.enums)
            all_enums.extend(command.response.enums)

        unique_enums = {}
        inconsistent_enums = []
        for maybe in all_enums:
            is_unique = True
            for unique_enum in unique_enums.values():
                if unique_enum == maybe:
                    is_unique = False
                    break

            if is_unique:
                # -- search for inconsistent duplicates
                if unique_enums.get(maybe.name):
                    existing = unique_enums[maybe.name]

                    inconsistent_enums.append({
                        'enum.0': {
                            'name': existing.name,
                            'values': list(existing.values),
                        },
                        'enum.1': {
                            'name': maybe.name,
                            'values': list(maybe.values),
                        },
                    })

                else:
                    unique_enums[maybe.name] = maybe

        if inconsistent_enums:
            raise self.ServerError(
                'INCONSISTENT_ENUMS_DETECTED',
                data={'inconsistent_enums': inconsistent_enums})

        return sorted(
            unique_enums.values(),
            key=lambda x: (x.name, x.index))

    #
    # API Service
    #
    def render_api_index_ts(self, commands_by_domain):

        blocks = []
        for domain in sorted(commands_by_domain.keys(), key=lambda x: x.id):
            blocks.append(
                "export * from './{domain.id}/index';".format(domain=domain))

        path = os.path.join(
            self.repo.base_path,
            'projects/client/src/domains/index.ts')
        with open(path, 'w') as f:
            f.write('\n'.join(blocks))

    def render_api_ts(self, commands_by_domain):

        blocks = [
            normalize_indentation('''
                {autogenerated_doc}

                /**
                 * Facade API Service for all domains
                 */
                import {{ Injectable, Injector }} from '@angular/core';
                import {{ Observable }} from 'rxjs';

                import * as X from '../domains/index';

                @Injectable()
                export class APIService {{

                    constructor(private injector: Injector) {{}}

            ''', 0).format(autogenerated_doc=self.AUTOGENERATED_DOC,)
        ]

        for domain in sorted(commands_by_domain.keys(), key=lambda x: x.id):
            commands = commands_by_domain[domain]
            blocks.append(normalize_indentation('''
                /**
                 * {domain.name} domain
                 */
                private _{domain.id}Domain: X.{domain.camel_id}Domain;

                public get {domain.id}Domain(): X.{domain.camel_id}Domain {{
                    if (!this._{domain.id}Domain) {{
                        this._{domain.id}Domain = this.injector.get(X.{domain.camel_id}Domain);
                    }}

                    return this._{domain.id}Domain;
                }}
            ''', 4).format(  # noqa
                domain=domain,
            ))

            for command_name in sorted(commands.keys()):
                command = commands[command_name]
                blocks.append(
                    normalize_indentation(command.render_facade(), 4))

        blocks.append('}')

        path = os.path.join(
            self.repo.base_path,
            'projects/client/src/services/api.service.ts')
        with open(path, 'w') as f:
            f.write('\n\n'.join(blocks))

    #
    # Client Module
    #
    def render_client_module_ts(self, commands_by_domain):

        domains_imports = []
        domains_list = []
        for domain in sorted(commands_by_domain.keys(), key=lambda x: x.id):
            domains_imports.append(normalize_indentation('''
                import {{ {domain.camel_id}Domain }} from './domains/{domain.id}/index';
            ''', 0).format(  # noqa
                domain=domain))
            domains_list.append(normalize_indentation('''
                    {domain.camel_id}Domain,
                ''', 8).format(domain=domain))

        blocks = [
            normalize_indentation('''
                {autogenerated_doc}

                import {{ NgModule, ModuleWithProviders }} from '@angular/core';
                import {{ HttpClientModule }} from '@angular/common/http';

                /** Domains */
                {domains_imports}

                /** Services */
                import {{ APIService }} from './services/api.service';
                import {{ HttpService, Config }} from './http.service';

                @NgModule({{
                    imports: [HttpClientModule],
                    providers: [
                        // Domains
                {domains_list}

                        // Facade
                        APIService,
                        HttpService,
                    ]
                }})
                export class ClientModule {{
                  static forRoot(config: Config): ModuleWithProviders<ClientModule> {{

                    return {{

                      ngModule: ClientModule,

                      providers: [{{ provide: 'COSPHERE_CLIENT_CONFIG_TOKEN', useValue: config }}]

                    }};
                  }}
                }}
            ''', 0).format(  # noqa
                autogenerated_doc=self.AUTOGENERATED_DOC,
                domains_imports='\n'.join(domains_imports),
                domains_list='\n'.join(domains_list))
        ]

        path = os.path.join(
            self.repo.base_path,
            'projects/client/src/client.module.ts')
        with open(path, 'w') as f:
            f.write('\n\n'.join(blocks))




import os
import json

from lily_assistant.config import Config
from lily.conf import settings
from django.core.management import CommandError
import djclick as click


@click.command()
def command():

    config = Config()

    commands_dir = os.path.join(config.get_lily_path(), 'commands')

    excluded = (
        settings.LILY_EXCLUDE_QUERY_PARSER_ALL_OPTIONAL_ASSERTIONS or [])
    with open(os.path.join(commands_dir, f'{config.version}.json'), 'r') as f:
        commands = json.loads(f.read())

        for name, command in commands.items():
            if name in excluded or name == '@enums':
                continue

            input_query = command['schemas'].get(
                'input_query',
                {'schema': {'required': []}})

            if input_query['schema']['required']:
                fields = ','.join(input_query['schema']['required'])

                raise CommandError(
                    f"ERROR: query parser for '{name}' has some not optional "
                    f"fields: [{fields}]")




from django.test import TestCase

from lily.docs.renderers.angular.domain import Path, Domain


class PathTestCase(TestCase):

    def test_join(self):
        assert Path('/home/jake').join('photo.png') == '/home/jake/photo.png'

    def test_join__with_extra_slash(self):
        assert Path('/home/jake').join('/photo.png') == '/home/jake/photo.png'


class DomainTestCase(TestCase):

    def test_attributes(self):

        d = Domain('cards', 'Cards Management')

        assert d.id == 'cards'
        assert d.name == 'Cards Management'
        assert d.camel_id == 'Cards'

    def test_path(self):

        path = Domain('cards', 'Cards Management').path

        assert isinstance(path, Path)
        assert path.base_path == './projects/client/src/domains/cards'




import os
import string

from itsdangerous import URLSafeTimedSerializer, SignatureExpired, BadData

from .events import EventFactory


event = EventFactory()


def sign_payload(email, payload, secret, salt):
    serializer = URLSafeTimedSerializer(secret)

    return serializer.dumps([email, payload], salt=salt)


def verify_payload(encoded_payload, secret, salt, signer_email, max_age):
    serializer = URLSafeTimedSerializer(secret)
    context = event.Context(email=signer_email)

    try:
        email, payload = serializer.loads(
            encoded_payload, salt=salt, max_age=max_age)

    except SignatureExpired:
        raise event.BrokenRequest(
            'PAYLOAD_VERIFIED_AS_EXPIRED', context=context)

    except BadData:
        raise event.BrokenRequest(
            'PAYLOAD_VERIFIED_AS_BROKEN', context=context)

    else:
        if email != signer_email:
            raise event.BrokenRequest(
                'PAYLOAD_VERIFIED_AS_BROKEN_MISMATCHING_EMAILS',
                context=context)

    return payload


def create_secret(length=64, with_punctuation=True):
    """Generate a secret stored on each User's Account.

    The values generated by this function resulted in 100% score on
    http://www.passwordmeter.com/.

    """
    chars = (
        string.ascii_uppercase +
        # introduce digits twice to boost the probability of its
        # occurrence in the secret
        string.digits + string.digits +
        string.ascii_lowercase)
    if with_punctuation:
        chars += string.punctuation

    chars_len = len(chars)
    secret = ''

    for i in range(length):
        secret += chars[ord(os.urandom(1)) % chars_len]

    return secret




from lily.base import serializers
from lily.base.meta import MetaSerializer
from lily.base.source import SourceSerializer
from lily.base.access import AccessSerializer


class CommandSerializer(serializers.Serializer):

    _type = 'command'

    method = serializers.EnumChoiceField(
        enum_name='command_method',
        choices=('POST', 'GET', 'PUT', 'DELETE'))

    path_conf = serializers.JSONField()

    meta = MetaSerializer()

    access = AccessSerializer()

    source = SourceSerializer()

    schemas = serializers.JSONField(required=False)

    examples = serializers.JSONField(required=False)




from unittest.mock import call
import asyncio

from django.test import TestCase
import pytest
import requests
from requests.exceptions import ConnectionError, Timeout

from lily.async import AsyncTask, BackoffExecutor


class Client:

    def task(self, task_id):

        return requests.get('http://hello.world/tasks/{}'.format(task_id))


class Response:

    def __init__(self, status_code, body={}):
        self.status_code = status_code
        self.body = body

    def __eq__(self, other):
        return (
            self.status_code == other.status_code and
            self.body == other.body)


async def async_sleep(*args):
    pass


class BackoffExecutorTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker):
        self.mocker = mocker

    def test_single_request__all_successful(self):

        get = self.mocker.patch.object(requests, 'get')
        get.side_effect = [
            Response(200),
        ]

        responses = BackoffExecutor([
            AsyncTask(callback=Client().task, args=[11])
        ]).run()

        assert responses == [Response(200)]
        assert get.call_args_list == [call('http://hello.world/tasks/11')]

    def test_multiple_requests__all_successful(self):

        get = self.mocker.patch.object(requests, 'get')
        get.side_effect = [
            Response(200),
            Response(201),
            Response(403),
        ]

        responses = BackoffExecutor([
            AsyncTask(callback=Client().task, args=[11]),
            AsyncTask(callback=Client().task, args=[14]),
            AsyncTask(callback=Client().task, args=[16]),
        ]).run()

        assert responses == [
            Response(200),
            Response(201),
            Response(403),
        ]
        assert get.call_args_list == [
            call('http://hello.world/tasks/11'),
            call('http://hello.world/tasks/14'),
            call('http://hello.world/tasks/16'),
        ]

    def test_multiple_requests__some_successful__backoff(self):

        sleep = self.mocker.patch.object(asyncio, 'sleep')
        sleep.side_effect = async_sleep

        get = self.mocker.patch.object(requests, 'get')
        get.side_effect = [
            # 1st. round
            Response(200),    # 1.
            Exception,        # 2.
            ConnectionError,  # 3.

            # 2nd. round
            Timeout,          # 2.
            Response(201),    # 3.

            # 3rd. round
            Response(403),    # 2.
        ]

        responses = BackoffExecutor([
            AsyncTask(callback=Client().task, args=[11]),
            AsyncTask(callback=Client().task, args=[14]),
            AsyncTask(callback=Client().task, args=[16]),
        ]).run()

        assert responses == [
            Response(200),
            Response(403),
            Response(201),
        ]
        assert sleep.call_args_list == [call(0), call(2), call(8)]
        assert get.call_args_list == [
            # 1st. round
            call('http://hello.world/tasks/11'),
            call('http://hello.world/tasks/14'),
            call('http://hello.world/tasks/16'),

            # 2nd. round
            call('http://hello.world/tasks/14'),
            call('http://hello.world/tasks/16'),

            # 3rd. round
            call('http://hello.world/tasks/14'),
        ]

    def test_multiple_requests__till_backoff_exhaustion(self):

        sleep = self.mocker.patch.object(asyncio, 'sleep')
        sleep.side_effect = async_sleep

        get = self.mocker.patch.object(requests, 'get')
        get.side_effect = [
            # 1st. round
            Response(200),    # 1.
            Exception,        # 2.
            ConnectionError,  # 3.

            # 2nd. round
            Timeout,          # 2.
            Response(201),    # 3.

            # 3rd. round
            Response(403),    # 2.
        ]

        responses = BackoffExecutor(
            [
                AsyncTask(callback=Client().task, args=[11]),
                AsyncTask(callback=Client().task, args=[14]),
                AsyncTask(callback=Client().task, args=[16]),
            ],
            max_attempts=2).run()

        assert len(responses) == 3
        assert responses[0] == Response(200)
        assert isinstance(responses[1], Timeout) is True

        assert sleep.call_args_list == [call(0), call(2)]
        assert get.call_args_list == [
            # 1st. round
            call('http://hello.world/tasks/11'),
            call('http://hello.world/tasks/14'),
            call('http://hello.world/tasks/16'),

            # 2nd. round
            call('http://hello.world/tasks/14'),
            call('http://hello.world/tasks/16'),
        ]




import hashlib
import hmac
from datetime import datetime

from django.http import HttpResponse
from django.conf import settings
from .input import Input
from .meta import Meta, Domain
from .command import command, HTTPCommands
from . import name, parsers


# FIXME: to be removed!!!!!
# it's not needed any more!!!! --> to cosphere specific
# it's still used in the game-service
class S3UploadSignCommands(HTTPCommands):

    class QueryParser(parsers.Parser):

        to_sign = parsers.CharField()

        datetime = parsers.CharField()

        def validate_datetime(self, value):
            try:
                return datetime.strptime(
                    value, r'%Y%m%dT%H%M%SZ').strftime(r'%Y%m%d')

            except ValueError:
                raise parsers.ValidationError(
                    'invalid datetime format accepted is YYYYMMDDThhmmssZ')

    @command(
        name=name.Execute('SIGN', 'S3_UPLOAD'),

        meta=Meta(
            title='Sign S3 File Upload',
            domain=Domain(id='utils', name='utils')),

        input=Input(query_parser=QueryParser),
    )
    def get(self, request):

        to_sign = request.input.query['to_sign']
        dt = request.input.query['datetime']

        signing_key = self.get_signature_key(
            settings.AWS_S3_ACCESS_KEY,
            dt,
            settings.AWS_S3_REGION,
            service='s3')

        return HttpResponse(self.get_signature(signing_key, to_sign))

    def get_signature_key(self, key, date_stamp, region, service):
        k_date = self.sign(('AWS4' + key).encode('utf-8'), date_stamp)
        k_region = self.sign(k_date, region)
        k_service = self.sign(k_region, service)
        k_signing = self.sign(k_service, 'aws4_request')

        return k_signing

    def sign(self, key, msg):

        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    def get_signature(self, signing_key, to_sign):

        return hmac.new(
            signing_key,
            to_sign.encode("utf-8"),
            hashlib.sha256).hexdigest()




from django.core.exceptions import ImproperlyConfigured


try:
    from .base.access import *  # noqa
    from .base.meta import *  # noqa
    from .base.source import *  # noqa
    from .base.input import *  # noqa
    from .base.output import *  # noqa
    from .base.command import *  # noqa
    from .base.authorizer import *  # noqa
    from .base import name, parsers, serializers, commands  # noqa

except ImproperlyConfigured:
    pass




import json
import os

from django.test import TestCase
from click.testing import CliRunner
import pytest

from lily.base.test import override_settings
from lily.assertion.management.commands.assert_query_parser_fields_are_optional import command  # noqa


class AssertQueryParserFieldsAreOptionalCommandTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()

    #
    # ASSERT_QUERY_PARSER_FIELDS_ARE_OPTIONAL
    #
    def test_assert_query_parser_fields_are_optional(self):

        self.mocker.patch.object(os, 'getcwd').return_value = str(self.tmpdir)
        lily_dir = self.tmpdir.mkdir('.lily')
        lily_dir.join('config.json').write(json.dumps({'version': '9.1.4'}))

        commands_dir = lily_dir.mkdir('commands')
        commands_dir.join('9.1.4.json').write(json.dumps({
            'BULK_READ_WHAT': {
                'schemas': {
                    'input_query': {
                        'schema': {
                            'required': [],
                        }
                    }
                }
            },
            'DELETE_IT': {
                'schemas': {
                }
            },
        }))

        result = self.runner.invoke(command)

        assert result.exit_code == 0
        assert result.output.strip() == ''

    def test_assert_query_parser_fields_are_optional__broken_detected(self):

        self.mocker.patch.object(os, 'getcwd').return_value = str(self.tmpdir)
        lily_dir = self.tmpdir.mkdir('.lily')
        lily_dir.join('config.json').write(json.dumps({'version': '9.1.4'}))

        commands_dir = lily_dir.mkdir('commands')
        commands_dir.join('9.1.4.json').write(json.dumps({
            'BULK_READ_WHAT': {
                'schemas': {
                    'input_query': {
                        'schema': {
                            'required': ['ids'],
                        }
                    }
                }
            },
            'DELETE_IT': {
                'schemas': {}
            },
        }))

        result = self.runner.invoke(command)

        assert result.exit_code == 1
        assert result.output.strip() == (
            "CommandError: ERROR: query parser for 'BULK_READ_WHAT' has some "
            "not optional fields: [ids]")

    @override_settings(LILY_EXCLUDE_QUERY_PARSER_ALL_OPTIONAL_ASSERTIONS=[
        'BULK_READ_WHAT',
    ])
    def test_assert_query_parser_fields_are_optional__excluded(self):

        self.mocker.patch.object(os, 'getcwd').return_value = str(self.tmpdir)
        lily_dir = self.tmpdir.mkdir('.lily')
        lily_dir.join('config.json').write(json.dumps({'version': '9.1.4'}))

        commands_dir = lily_dir.mkdir('commands')
        commands_dir.join('9.1.4.json').write(json.dumps({
            'BULK_READ_WHAT': {
                'schemas': {
                    'input_query': {
                        'schema': {
                            'required': ['ids'],
                        }
                    }
                }
            },
            'DELETE_IT': {
                'schemas': {}
            },
        }))

        result = self.runner.invoke(command)

        assert result.exit_code == 0
        assert result.output.strip() == ''




import re


def capitalize(t):

    if len(t) > 1:
        return t[0].upper() + t[1:]

    return t.upper()


def to_camelcase(text, first_lower=False):

    if text == text.upper():
        text = text.lower()

    tokens = re.split(r'\_+', text)

    camel_text = ''.join([capitalize(token) for token in tokens])

    if camel_text and first_lower:
        camel_text = camel_text[0].lower() + camel_text[1:]

    return camel_text




import os
import json


class StopWordsFilter:

    def __init__(self):

        base_dir = os.path.dirname(__file__)

        self.stopwords = {}

        # -- polish
        with open(os.path.join(base_dir, 'pl.json'), 'r') as f:
            self.stopwords['polish'] = set(json.loads(f.read()))

    def is_stopword(self, conf, word):
        if conf == 'polish':
            return word in self.stopwords[conf]

        # -- all other languages are treated the same
        return False


stopwords_filter = StopWordsFilter()




import textwrap

import click


class Logger:

    def info(self, text):

        text = textwrap.dedent(text).strip()
        click.secho('[INFO]\n\n{text}'.format(text=text), fg='yellow')

    def error(self, text):

        text = textwrap.dedent(text).strip()
        click.secho('[ERROR]\n\n{text}'.format(text=text), fg='red')




import os
import json
from copy import deepcopy
from unittest.mock import patch
from contextlib import ContextDecorator

from django.urls import get_resolver
from django.test import Client as DjangoClient
from lily_assistant.config import Config

from lily.conf import settings


class MissingConfError(Exception):
    pass


def get_examples_filepath():
    return os.path.join(Config.get_lily_path(), 'examples.json')


class override_settings(ContextDecorator):  # noqa

    def __init__(self, **settings):
        self.settings = settings
        self.patchers = []

    def __enter__(self):

        self.patchers = [
            patch.object(settings, key, value)
            for key, value in self.settings.items()
        ]
        for p in self.patchers:
            p.start()

        return self

    def __exit__(self, exc_type, exc, exc_tb):
        for p in self.patchers:
            p.stop()


class Client(DjangoClient):

    resolver = get_resolver(None)

    def post(self, *args, **kwargs):
        return self._make_request('post', *args, **kwargs)

    def get(self, *args, **kwargs):
        return self._make_request('get', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._make_request('put', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._make_request('delete', *args, **kwargs)

    def _make_request(self, http_verb, *args, **kwargs):

        def assure_path_exists(path):
            dir_path = os.path.dirname(path)
            if not os.path.exists(dir_path):
                os.makedirs(dir_path)

        assure_path_exists(get_examples_filepath())

        response = getattr(super(Client, self), http_verb)(*args, **kwargs)

        fn = self.resolver.resolve(response.request['PATH_INFO']).func

        try:
            command_conf = getattr(fn.view_class, http_verb).command_conf

            try:
                with open(get_examples_filepath(), 'r+') as f:
                    examples = json.loads(f.read() or '{}')

            except FileNotFoundError:
                examples = {}

            command_name = command_conf['name']
            examples.setdefault(command_name, {})

            try:
                response_content = response.json()

            except ValueError:
                response_content = str(response.content, encoding='utf-8')
                event = None

            else:
                event = response_content['@event']

            # -- render example key name
            try:
                extra_desc = kwargs.pop('extra_desc')

            except KeyError:
                if event:
                    example_key = f'{response.status_code} ({event})'  # noqa

                else:
                    example_key = f'{response.status_code}'

            else:
                if event:
                    example_key = (
                        f'{response.status_code} ({event}) - {extra_desc}')

                else:
                    example_key = f'{response.status_code} - {extra_desc}'

            # -- construct the request
            request = {}
            path = response.request['PATH_INFO']

            if response.request.get('QUERY_STRING'):
                path = '{path}?{query_string}'.format(
                    path=path,
                    query_string=response.request['QUERY_STRING'])

            request['path'] = path
            request_kwargs = deepcopy(kwargs)
            data = request_kwargs.get('data')
            if data:
                data = request_kwargs.pop('data')

                if http_verb.upper() in ['POST', 'PUT']:
                    if isinstance(data, dict):
                        request['content'] = data

                    else:
                        request['content'] = json.loads(data)

            if request_kwargs:
                def normalize_header(h):
                    return h.upper().replace('_', '-').replace('HTTP-', '')

                request['headers'] = {
                    normalize_header(h): v
                    for h, v in request_kwargs.items()}

            examples[command_name][example_key] = {
                'method': http_verb,
                'description': event,
                'request': request,
                'response': {
                    'status': response.status_code,
                    'content_type': response['Content-Type'],
                    'content': response_content,
                },
            }

            with open(get_examples_filepath(), 'w') as f:
                f.write(json.dumps(examples, indent=4))

        except AttributeError:
            raise MissingConfError({
                'method': http_verb,
                'path': response.request['PATH_INFO'],
            })

        return response




import inspect

from lily_assistant.config import Config

from . import serializers


class Source:

    def __init__(self, fn):
        code, firstline = inspect.getsourcelines(fn)
        self.filepath = inspect.getfile(fn).replace(
            Config.get_project_path(), '')
        self.start_line = firstline
        self.end_line = firstline + len(code) - 1

    def __eq__(self, other):
        return (
            self.filepath == other.filepath and
            self.start_line == other.start_line and
            self.end_line == other.end_line)


class SourceSerializer(serializers.Serializer):

    _type = 'source'

    filepath = serializers.CharField()

    start_line = serializers.IntegerField()

    end_line = serializers.IntegerField()




import os
from unittest import TestCase
import json

from click.testing import CliRunner
import pytest

from lily.base.management.commands.clear_examples import command


class ClearExamplesCommandTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()

    #
    # CLEAR_EXAMPLES
    #
    def test_clear_examples__file_exists(self):

        self.mocker.patch.object(os, 'getcwd').return_value = str(self.tmpdir)
        lily_dir = self.tmpdir.mkdir('.lily')
        lily_dir.join('examples.json').write(json.dumps({'some': 'examples'}))

        result = self.runner.invoke(command)

        assert result.exit_code == 0
        assert result.output.strip() == "'examples.json' was removed"
        assert lily_dir.listdir() == []

    def test_clear_examples__file_does_not_exist(self):

        self.mocker.patch.object(os, 'getcwd').return_value = str(self.tmpdir)
        lily_dir = self.tmpdir.mkdir('.lily')

        result = self.runner.invoke(command)

        assert result.exit_code == 0
        assert result.output.strip() == "'examples.json' was removed"
        assert lily_dir.listdir() == []
